# SPDX-License-Identifier: MIT
# Source: https://github.com/microsoft/MaskFlownet/tree/5cba12772e2201f0d1c1e27161d224e585334571
from mxnet.gluon import nn
import math

from . import layer

class Downsample(nn.HybridBlock):
	def __init__(self, factor, **kwargs):
		super().__init__(**kwargs)
		self.factor = factor

	@staticmethod
	def _kernel2d(F, w):
		kernel = ((w + 1) - F.abs((w - F.arange(0, w * 2 + 1)))) / (2 * w + 1)
		kernel = F.broadcast_mul(F.expand_dims(kernel, axis=0), F.expand_dims(kernel, axis=1))
		return F.reshape(kernel, (1, 1, w * 2 + 1, w * 2 + 1))

	def hybrid_forward(self, F, img):
		if self.factor == 1:
			return img
		batch_img = F.expand_dims(F.reshape(img, [-3, -2]), axis=1)
		factor=self.factor
		kernel = self._kernel2d(F, factor // 2)
		conv_args = dict(
			weight=kernel, 
			no_bias=True, 
			kernel=(factor + 1,) * 2,
			stride=(factor,) * 2,
			pad=(factor // 2,) * 2,
			num_filter=1)
		upsamp_nom = F.Convolution(data=F.ones_like(batch_img), **conv_args)
		upsamp_img = F.Convolution(data=batch_img, **conv_args)
		return F.broadcast_div(F.reshape_like(upsamp_img, img, lhs_begin = 0, lhs_end = 2, rhs_begin = 0, rhs_end = 2),
			F.reshape_like(upsamp_nom, img, lhs_begin = 0, lhs_end = 2, rhs_begin = 0, rhs_end = 2))

class Upsample(nn.HybridBlock):
	def __init__(self, factor, **kwargs):
		super().__init__(**kwargs)
		self.factor = factor

	@staticmethod
	def _kernel2d(F, w):
		c = w // 2
		kernel = 1 - F.abs(c - F.arange(0, w)) / (c + 1)
		kernel = F.broadcast_mul(F.expand_dims(kernel, axis=0), F.expand_dims(kernel, axis=1))
		return F.reshape(kernel, (1, 1, w, w))

	def hybrid_forward(self, F, img):
		if self.factor == 1:
			return img
		batch_img = F.expand_dims(F.reshape(img, [-3, -2]), axis=1)
		batch_img = F.pad(batch_img, mode = "edge", pad_width = (0, 0, 0, 0, 0, 1, 0, 1))
		factor = self.factor
		kernel = self._kernel2d(F, factor * 2 - 1)
		conv_args = dict(
			weight=kernel, 
			no_bias=True, 
			kernel=(factor * 2 - 1,) * 2,
			stride=(factor,) * 2,
			pad=(factor - 1,) * 2,
			num_filter=1)
		upsamp_img = F.slice(F.Deconvolution(data=batch_img, **conv_args), begin=(None, None, None, None), end=(None, None, -1, -1))
		return F.reshape_like(upsamp_img, img, lhs_begin = 0, lhs_end = 2, rhs_begin = 0, rhs_end = 2)

use_bias = True

class MaskFlownet_S(nn.HybridBlock):
	def __init__(self, config = None, **kwargs):
		super().__init__(**kwargs)
		self.scale = 20. * config.network.flow_multiplier.get(1.)
		self.md = 4
		self.strides = [64, 32, 16, 8, 4]
		self.deform_bias = config.network.deform_bias.get(True)
		self.upfeat_ch = config.network.upfeat_ch.get([16, 16, 16, 16])
		
		with self.name_scope():
			self.activate = nn.LeakyReLU(0.1)
			self.warp = layer.Reconstruction2D(2)
			
			self.conv1a = self.conv(  3,  16, kernel_size=3, stride=2, padding=1, dilation=1, prefix = 'conv1a')
			self.conv1b = self.conv( 16,  16, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv1b')
			self.conv1c = self.conv( 16,  16, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv1c')
			self.conv2a = self.conv( 16,  32, kernel_size=3, stride=2, padding=1, dilation=1, prefix = 'conv2a')
			self.conv2b = self.conv( 32,  32, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv2b')
			self.conv2c = self.conv( 32,  32, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv2c')
			self.conv3a = self.conv( 32,  64, kernel_size=3, stride=2, padding=1, dilation=1, prefix = 'conv3a')
			self.conv3b = self.conv( 64,  64, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv3b')
			self.conv3c = self.conv( 64,  64, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv3c')
			self.conv4a = self.conv( 64,  96, kernel_size=3, stride=2, padding=1, dilation=1, prefix = 'conv4a')
			self.conv4b = self.conv( 96,  96, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv4b')
			self.conv4c = self.conv( 96,  96, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv4c')
			self.conv5a = self.conv( 96, 128, kernel_size=3, stride=2, padding=1, dilation=1, prefix = 'conv5a')
			self.conv5b = self.conv(128, 128, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv5b')
			self.conv5c = self.conv(128, 128, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv5c')
			self.conv6a = self.conv(128, 196, kernel_size=3, stride=2, padding=1, dilation=1, prefix = 'conv6a')
			self.conv6b = self.conv(196, 196, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv6b')
			self.conv6c = self.conv(196, 196, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv6c')

			self.leakyRELU = nn.LeakyReLU(0.1)
			ch = None

			self.conv6_0 = self.conv(ch, 128, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv6_0')
			self.conv6_1 = self.conv(ch, 128, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv6_1')
			self.conv6_2 = self.conv(ch,  96, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv6_2')
			self.conv6_3 = self.conv(ch,  64, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv6_3')
			self.conv6_4 = self.conv(ch,  32, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv6_4')

			self.conv5_0 = self.conv(ch, 128, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv5_0')
			self.conv5_1 = self.conv(ch, 128, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv5_1')
			self.conv5_2 = self.conv(ch,  96, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv5_2')
			self.conv5_3 = self.conv(ch,  64, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv5_3')
			self.conv5_4 = self.conv(ch,  32, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv5_4')
			
			self.conv4_0 = self.conv(ch, 128, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv4_0')
			self.conv4_1 = self.conv(ch, 128, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv4_1')
			self.conv4_2 = self.conv(ch,  96, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv4_2')
			self.conv4_3 = self.conv(ch,  64, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv4_3')
			self.conv4_4 = self.conv(ch,  32, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv4_4')
			
			self.conv3_0 = self.conv(ch, 128, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv3_0')
			self.conv3_1 = self.conv(ch, 128, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv3_1')
			self.conv3_2 = self.conv(ch,  96, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv3_2')
			self.conv3_3 = self.conv(ch,  64, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv3_3')
			self.conv3_4 = self.conv(ch,  32, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv3_4')
			
			self.conv2_0 = self.conv(ch, 128, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv2_0')
			self.conv2_1 = self.conv(ch, 128, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv2_1')
			self.conv2_2 = self.conv(ch,  96, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv2_2')
			self.conv2_3 = self.conv(ch,  64, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv2_3')
			self.conv2_4 = self.conv(ch,  32, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv2_4')
			
			self.dc_conv1 = self.conv(ch, 128, kernel_size=3, stride=1, padding=1,  dilation=1,  prefix = 'dc_conv1')
			self.dc_conv2 = self.conv(ch, 128, kernel_size=3, stride=1, padding=2,  dilation=2,  prefix = 'dc_conv2')
			self.dc_conv3 = self.conv(ch, 128, kernel_size=3, stride=1, padding=4,  dilation=4,  prefix = 'dc_conv3')
			self.dc_conv4 = self.conv(ch,  96, kernel_size=3, stride=1, padding=8,  dilation=8,  prefix = 'dc_conv4')
			self.dc_conv5 = self.conv(ch,  64, kernel_size=3, stride=1, padding=16, dilation=16, prefix = 'dc_conv5')
			self.dc_conv6 = self.conv(ch,  32, kernel_size=3, stride=1, padding=1,  dilation=1,  prefix = 'dc_conv6')
			self.dc_conv7 = self.predict_flow(ch, prefix = 'dc_conv7')

			self.upfeat5 = self.deconv(ch, self.upfeat_ch[0], kernel_size=4, stride=2, padding=1, prefix = 'upfeat5')
			self.upfeat4 = self.deconv(ch, self.upfeat_ch[1], kernel_size=4, stride=2, padding=1, prefix = 'upfeat4')
			self.upfeat3 = self.deconv(ch, self.upfeat_ch[2], kernel_size=4, stride=2, padding=1, prefix = 'upfeat3')
			self.upfeat2 = self.deconv(ch, self.upfeat_ch[3], kernel_size=4, stride=2, padding=1, prefix = 'upfeat2')

			self.pred_flow6 = self.predict_flow(ch, prefix = 'pred_flow6')
			self.pred_flow5 = self.predict_flow(ch, prefix = 'pred_flow5')
			self.pred_flow4 = self.predict_flow(ch, prefix = 'pred_flow4')
			self.pred_flow3 = self.predict_flow(ch, prefix = 'pred_flow3')
			self.pred_flow2 = self.predict_flow(ch, prefix = 'pred_flow2')

			self.pred_mask6 = self.predict_mask(ch, prefix = 'pred_mask6')
			self.pred_mask5 = self.predict_mask(ch, prefix = 'pred_mask5')
			self.pred_mask4 = self.predict_mask(ch, prefix = 'pred_mask4')
			self.pred_mask3 = self.predict_mask(ch, prefix = 'pred_mask3')

			self.deform5 = layer.DeformableConv2D(128, kernel_size=3, strides=1, padding=1, use_bias=self.deform_bias, prefix = 'deform5')
			self.deform4 = layer.DeformableConv2D( 96, kernel_size=3, strides=1, padding=1, use_bias=self.deform_bias, prefix = 'deform4')
			self.deform3 = layer.DeformableConv2D( 64, kernel_size=3, strides=1, padding=1, use_bias=self.deform_bias, prefix = 'deform3')
			self.deform2 = layer.DeformableConv2D( 32, kernel_size=3, strides=1, padding=1, use_bias=self.deform_bias, prefix = 'deform2')

			self.conv5f = self.conv(ch, 128, kernel_size = 3, stride = 1, activation = False, prefix = 'conv5f')
			self.conv4f = self.conv(ch,  96, kernel_size = 3, stride = 1, activation = False, prefix = 'conv4f')
			self.conv3f = self.conv(ch,  64, kernel_size = 3, stride = 1, activation = False, prefix = 'conv3f')
			self.conv2f = self.conv(ch,  32, kernel_size = 3, stride = 1, activation = False, prefix = 'conv2f')
				

	def conv(self, _, channels, kernel_size = 3, stride = 1, padding = 1, dilation = 1, activation = True, prefix = None):

		net = nn.HybridSequential()
		with net.name_scope():
			net.add(nn.Conv2D(channels = channels, kernel_size = kernel_size, strides = stride, padding = padding, dilation = dilation, prefix = prefix))
			if activation:
				net.add(self.activate)

		return net

	def deconv(self, _, channels, kernel_size = 4, stride = 2, padding = 1, prefix = None):

		net = nn.HybridSequential()
		with net.name_scope():
			net.add(nn.Conv2DTranspose(channels = channels, kernel_size = kernel_size, strides = stride, padding = padding, prefix = prefix))
			net.add(self.activate)

		return net

	def predict_flow(self, _, kernel_size = 3, stride = 1, padding = 1, prefix = None):

		return nn.Conv2D(2, kernel_size = kernel_size, strides = stride, padding = padding, prefix = prefix)

	def predict_mask(self, _, kernel_size = 3, stride = 1, padding = 1, prefix = None):

		return nn.Conv2D(1, kernel_size = kernel_size, strides = stride, padding = padding, prefix = prefix)

	def corr(self, F, im1, im2, stride1 = 1, stride2 = 1):

		return F.Correlation(im1, im2, pad_size = self.md, kernel_size = 1, max_displacement = self.md, stride1 = stride1, stride2 = stride2, is_multiply = 1)

	def hybrid_forward(self, F, im1, im2):
		c10 = im1
		c20 = im2

		c11 = self.conv1c(self.conv1b(self.conv1a(c10)))
		c21 = self.conv1c(self.conv1b(self.conv1a(c20)))
		c12 = self.conv2c(self.conv2b(self.conv2a(c11)))
		c22 = self.conv2c(self.conv2b(self.conv2a(c21)))
		c13 = self.conv3c(self.conv3b(self.conv3a(c12)))
		c23 = self.conv3c(self.conv3b(self.conv3a(c22)))
		c14 = self.conv4c(self.conv4b(self.conv4a(c13)))
		c24 = self.conv4c(self.conv4b(self.conv4a(c23)))
		c15 = self.conv5c(self.conv5b(self.conv5a(c14)))
		c25 = self.conv5c(self.conv5b(self.conv5a(c24)))
		c16 = self.conv6c(self.conv6b(self.conv6a(c15)))
		c26 = self.conv6c(self.conv6b(self.conv6a(c25)))


		warp6 = c26
		corr6 = self.corr(F, c16, warp6)
		corr6 = self.leakyRELU(corr6)
		x = corr6
		x = F.concat(self.conv6_0(x), x, dim=1)
		x = F.concat(self.conv6_1(x), x, dim=1)
		x = F.concat(self.conv6_2(x), x, dim=1)
		x = F.concat(self.conv6_3(x), x, dim=1)
		x = F.concat(self.conv6_4(x), x, dim=1)
		flow6 = self.pred_flow6(x)
		mask6 = self.pred_mask6(x)

		feat5 = self.upfeat5(x)
		flow5 = Upsample(2)(flow6)
		mask5 = Upsample(2)(mask6)
		warp5 = self.deform5(c25, F.repeat(F.expand_dims(flow5*self.scale/self.strides[1], axis=1), 9, axis=1).reshape((0, -3, -2)))
		tradeoff5 = feat5
		warp5 = F.broadcast_mul(warp5, F.sigmoid(mask5)) + self.conv5f(tradeoff5)
		warp5 = self.leakyRELU(warp5)
		corr5 = self.corr(F, c15, warp5) 
		corr5 = self.leakyRELU(corr5)
		x = F.concat(corr5, c15, feat5, flow5, dim=1)
		x = F.concat(self.conv5_0(x), x, dim=1)
		x = F.concat(self.conv5_1(x), x, dim=1)
		x = F.concat(self.conv5_2(x), x, dim=1)
		x = F.concat(self.conv5_3(x), x, dim=1)
		x = F.concat(self.conv5_4(x), x, dim=1)
		flow5 = flow5 + self.pred_flow5(x)
		mask5 = self.pred_mask5(x)

		feat4 = self.upfeat4(x)
		flow4 = Upsample(2)(flow5)
		mask4 = Upsample(2)(mask5)
		warp4 = self.deform4(c24, F.repeat(F.expand_dims(flow4*self.scale/self.strides[2], axis=1), 9, axis=1).reshape((0, -3, -2)))
		tradeoff4 = feat4
		warp4 = F.broadcast_mul(warp4, F.sigmoid(mask4)) + self.conv4f(tradeoff4)
		warp4 = self.leakyRELU(warp4)
		corr4 = self.corr(F, c14, warp4) 
		corr4 = self.leakyRELU(corr4)
		x = F.concat(corr4, c14, feat4, flow4, dim=1)
		x = F.concat(self.conv4_0(x), x, dim=1)
		x = F.concat(self.conv4_1(x), x, dim=1)
		x = F.concat(self.conv4_2(x), x, dim=1)
		x = F.concat(self.conv4_3(x), x, dim=1)
		x = F.concat(self.conv4_4(x), x, dim=1)
		flow4 = flow4 + self.pred_flow4(x)
		mask4 = self.pred_mask4(x)

		feat3 = self.upfeat3(x)
		flow3 = Upsample(2)(flow4)
		mask3 = Upsample(2)(mask4)
		warp3 = self.deform3(c23, F.repeat(F.expand_dims(flow3*self.scale/self.strides[3], axis=1), 9, axis=1).reshape((0, -3, -2)))
		tradeoff3 = feat3
		warp3 = F.broadcast_mul(warp3, F.sigmoid(mask3)) + self.conv3f(tradeoff3)
		warp3 = self.leakyRELU(warp3)
		corr3 = self.corr(F, c13, warp3) 
		corr3 = self.leakyRELU(corr3)
		x = F.concat(corr3, c13, feat3, flow3, dim=1)
		x = F.concat(self.conv3_0(x), x, dim=1)
		x = F.concat(self.conv3_1(x), x, dim=1)
		x = F.concat(self.conv3_2(x), x, dim=1)
		x = F.concat(self.conv3_3(x), x, dim=1)
		x = F.concat(self.conv3_4(x), x, dim=1)
		flow3 = flow3 + self.pred_flow3(x)
		mask3 = self.pred_mask3(x)
		
		feat2 = self.upfeat2(x)
		flow2 = Upsample(2)(flow3)
		mask2 = Upsample(2)(mask3)
		warp2 = self.deform2(c22, F.repeat(F.expand_dims(flow2*self.scale/self.strides[4], axis=1), 9, axis=1).reshape((0, -3, -2)))
		tradeoff2 = feat2
		warp2 = F.broadcast_mul(warp2, F.sigmoid(mask2)) + self.conv2f(tradeoff2)
		warp2 = self.leakyRELU(warp2)
		corr2 = self.corr(F, c12, warp2) 
		corr2 = self.leakyRELU(corr2)
		x = F.concat(corr2, c12, feat2, flow2, dim=1)
		x = F.concat(self.conv2_0(x), x, dim=1)
		x = F.concat(self.conv2_1(x), x, dim=1)
		x = F.concat(self.conv2_2(x), x, dim=1)
		x = F.concat(self.conv2_3(x), x, dim=1)
		x = F.concat(self.conv2_4(x), x, dim=1)
		flow2 = flow2 + self.pred_flow2(x)


		x = self.dc_conv4(self.dc_conv3(self.dc_conv2(self.dc_conv1(x))))
		flow2 = flow2 + self.dc_conv7(self.dc_conv6(self.dc_conv5(x)))

		predictions = [flow * self.scale for flow in [flow6, flow5, flow4, flow3, flow2]]
		occlusion_masks = []
		occlusion_masks.append(F.sigmoid(mask2))
		c1s = [c11, c12, c13, c14, c15, c16]
		c2s = [c21, c12, c13, c24, c25, c26]
		flows = [flow6, flow5, flow4, flow3, flow2]
		mask0 = Upsample(4)(mask2)
		mask0 = F.sigmoid(mask0) - 0.5
		c30 = c10
		c40 = self.warp(c20, Upsample(4)(flow2)*self.scale)
		c30 = F.concat(c30, F.zeros_like(mask0), dim=1)
		c40 = F.concat(c40, mask0, dim=1)
		srcs = [c1s, c2s, flows, c30, c40]
		return predictions, occlusion_masks, srcs


class MaskFlownet(nn.HybridBlock):
	def __init__(self, config = None, **kwargs):
		super().__init__(**kwargs)
		self.strides = [64, 32, 16, 8, 4]
		self.md = 2
		self.scale = 20. * config.network.flow_multiplier.get(1.)
		self.deform_bias = config.network.deform_bias.get(True)
		self.upfeat_ch = config.network.upfeat_ch.get([16, 16, 16, 16])

		with self.name_scope():
			self.MaskFlownet_S = MaskFlownet_S(config)
			self.activate = nn.LeakyReLU(0.1)
			self.warp = layer.Reconstruction2D(2)

			self.conv1x = self.conv(  3,  16, kernel_size=3, stride=2, padding=1, dilation=1, prefix = 'conv1x')
			self.conv1y = self.conv( 16,  16, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv1y')
			self.conv1z = self.conv( 16,  16, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv1z')
			self.conv2x = self.conv( 16,  32, kernel_size=3, stride=2, padding=1, dilation=1, prefix = 'conv2x')
			self.conv2y = self.conv( 32,  32, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv2y')
			self.conv2z = self.conv( 32,  32, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv2z')
			self.conv3x = self.conv( 32,  64, kernel_size=3, stride=2, padding=1, dilation=1, prefix = 'conv3x')
			self.conv3y = self.conv( 64,  64, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv3y')
			self.conv3z = self.conv( 64,  64, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv3z')
			self.conv4x = self.conv( 64,  96, kernel_size=3, stride=2, padding=1, dilation=1, prefix = 'conv4x')
			self.conv4y = self.conv( 96,  96, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv4y')
			self.conv4z = self.conv( 96,  96, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv4z')
			self.conv5x = self.conv( 96, 128, kernel_size=3, stride=2, padding=1, dilation=1, prefix = 'conv5x')
			self.conv5y = self.conv(128, 128, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv5y')
			self.conv5z = self.conv(128, 128, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv5z')
			self.conv6x = self.conv(128, 196, kernel_size=3, stride=2, padding=1, dilation=1, prefix = 'conv6x')
			self.conv6y = self.conv(196, 196, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv6y')
			self.conv6z = self.conv(196, 196, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv6z')

			self.leakyRELU = nn.LeakyReLU(0.1)
			ch = None

			self.conv6_0 = self.conv(ch, 128, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv6_0')
			self.conv6_1 = self.conv(ch, 128, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv6_1')
			self.conv6_2 = self.conv(ch,  96, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv6_2')
			self.conv6_3 = self.conv(ch,  64, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv6_3')
			self.conv6_4 = self.conv(ch,  32, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv6_4')

			self.conv5_0 = self.conv(ch, 128, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv5_0')
			self.conv5_1 = self.conv(ch, 128, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv5_1')
			self.conv5_2 = self.conv(ch,  96, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv5_2')
			self.conv5_3 = self.conv(ch,  64, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv5_3')
			self.conv5_4 = self.conv(ch,  32, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv5_4')
			
			self.conv4_0 = self.conv(ch, 128, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv4_0')
			self.conv4_1 = self.conv(ch, 128, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv4_1')
			self.conv4_2 = self.conv(ch,  96, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv4_2')
			self.conv4_3 = self.conv(ch,  64, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv4_3')
			self.conv4_4 = self.conv(ch,  32, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv4_4')
			
			self.conv3_0 = self.conv(ch, 128, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv3_0')
			self.conv3_1 = self.conv(ch, 128, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv3_1')
			self.conv3_2 = self.conv(ch,  96, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv3_2')
			self.conv3_3 = self.conv(ch,  64, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv3_3')
			self.conv3_4 = self.conv(ch,  32, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv3_4')
			
			self.conv2_0 = self.conv(ch, 128, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv2_0')
			self.conv2_1 = self.conv(ch, 128, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv2_1')
			self.conv2_2 = self.conv(ch,  96, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv2_2')
			self.conv2_3 = self.conv(ch,  64, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv2_3')
			self.conv2_4 = self.conv(ch,  32, kernel_size=3, stride=1, padding=1, dilation=1, prefix = 'conv2_4')
			
			self.dc_conv1 = self.conv(ch, 128, kernel_size=3, stride=1, padding=1,  dilation=1,  prefix = 'dc_conv1')
			self.dc_conv2 = self.conv(ch, 128, kernel_size=3, stride=1, padding=2,  dilation=2,  prefix = 'dc_conv2')
			self.dc_conv3 = self.conv(ch, 128, kernel_size=3, stride=1, padding=4,  dilation=4,  prefix = 'dc_conv3')
			self.dc_conv4 = self.conv(ch,  96, kernel_size=3, stride=1, padding=8,  dilation=8,  prefix = 'dc_conv4')
			self.dc_conv5 = self.conv(ch,  64, kernel_size=3, stride=1, padding=16, dilation=16, prefix = 'dc_conv5')
			self.dc_conv6 = self.conv(ch,  32, kernel_size=3, stride=1, padding=1,  dilation=1,  prefix = 'dc_conv6')
			self.dc_conv7 = self.predict_flow(ch, prefix = 'dc_conv7')

			self.upfeat5 = self.deconv(ch, self.upfeat_ch[0], kernel_size=4, stride=2, padding=1, prefix = 'upfeat5')
			self.upfeat4 = self.deconv(ch, self.upfeat_ch[1], kernel_size=4, stride=2, padding=1, prefix = 'upfeat4')
			self.upfeat3 = self.deconv(ch, self.upfeat_ch[2], kernel_size=4, stride=2, padding=1, prefix = 'upfeat3')
			self.upfeat2 = self.deconv(ch, self.upfeat_ch[3], kernel_size=4, stride=2, padding=1, prefix = 'upfeat2')

			self.pred_flow6 = self.predict_flow(ch, prefix = 'pred_flow6')
			self.pred_flow5 = self.predict_flow(ch, prefix = 'pred_flow5')
			self.pred_flow4 = self.predict_flow(ch, prefix = 'pred_flow4')
			self.pred_flow3 = self.predict_flow(ch, prefix = 'pred_flow3')
			self.pred_flow2 = self.predict_flow(ch, prefix = 'pred_flow2')

			self.deform6 = layer.DeformableConv2D(196, kernel_size=3, strides=1, padding=1, use_bias=self.deform_bias, prefix = 'deform6')
			self.deform5 = layer.DeformableConv2D(128, kernel_size=3, strides=1, padding=1, use_bias=self.deform_bias, prefix = 'deform5')
			self.deform4 = layer.DeformableConv2D( 96, kernel_size=3, strides=1, padding=1, use_bias=self.deform_bias, prefix = 'deform4')
			self.deform3 = layer.DeformableConv2D( 64, kernel_size=3, strides=1, padding=1, use_bias=self.deform_bias, prefix = 'deform3')
			self.deform2 = layer.DeformableConv2D( 32, kernel_size=3, strides=1, padding=1, use_bias=self.deform_bias, prefix = 'deform2')
				

	def load_head(self, ckpt, ctx):
		self.MaskFlownet_S.load_parameters(ckpt, ctx)

	def fix_head(self):
		for _, w in self.MaskFlownet_S.collect_params().items():
			w.grad_req = 'null'

	def conv(self, _, channels, kernel_size = 3, stride = 1, padding = 1, dilation = 1, activation = True, prefix = None):
		net = nn.HybridSequential()
		with net.name_scope():
			net.add(nn.Conv2D(channels = channels, kernel_size = kernel_size, strides = stride, padding = padding, dilation = dilation, prefix = prefix))
			if activation:
				net.add(self.activate)

		return net

	def deconv(self, _, channels, kernel_size = 4, stride = 2, padding = 1, prefix = None):
		net = nn.HybridSequential()
		with net.name_scope():
			net.add(nn.Conv2DTranspose(channels = channels, kernel_size = kernel_size, strides = stride, padding = padding, prefix = prefix))
			net.add(self.activate)

		return net

	def predict_flow(self, _, kernel_size = 3, stride = 1, padding = 1, prefix = None):
		return nn.Conv2D(2, kernel_size = kernel_size, strides = stride, padding = padding, prefix = prefix)

	def predict_mask(self, _, kernel_size = 3, stride = 1, padding = 1, prefix = None):
		return nn.Conv2D(1, kernel_size = kernel_size, strides = stride, padding = padding, prefix = prefix)

	def corr(self, F, im1, im2, stride1 = 1, stride2 = 1):
		return F.Correlation(im1, im2, pad_size = self.md, kernel_size = 1, max_displacement = self.md, stride1 = stride1, stride2 = stride2, is_multiply = 1)

	def hybrid_forward(self, F, im1, im2):
		_, _, srcs = self.MaskFlownet_S(im1, im2)
		c1s, c2s, flows, c30, c40 = srcs
		c11, c12, c13, c14, c15, c16 = c1s
		c21, c22, c23, c24, c25, c26 = c2s
		
		c31 = self.conv1z(self.conv1y(self.conv1x(c30)))
		c32 = self.conv2z(self.conv2y(self.conv2x(c31)))
		c33 = self.conv3z(self.conv3y(self.conv3x(c32)))
		c34 = self.conv4z(self.conv4y(self.conv4x(c33)))
		c35 = self.conv5z(self.conv5y(self.conv5x(c34)))
		c36 = self.conv6z(self.conv6y(self.conv6x(c35)))

		c41 = self.conv1z(self.conv1y(self.conv1x(c40)))
		c42 = self.conv2z(self.conv2y(self.conv2x(c41)))
		c43 = self.conv3z(self.conv3y(self.conv3x(c42)))
		c44 = self.conv4z(self.conv4y(self.conv4x(c43)))
		c45 = self.conv5z(self.conv5y(self.conv5x(c44)))
		c46 = self.conv6z(self.conv6y(self.conv6x(c45)))


		flow6 = flows[0]
		warp6u = self.deform6(c26, F.repeat(F.expand_dims(flow6*self.scale/self.strides[0], axis=1), 9, axis=1).reshape((0, -3, -2)))
		warp6u = self.leakyRELU(warp6u)
		corr6u = self.leakyRELU(self.corr(F, c16, warp6u))
		warp6v = c46
		corr6v = self.leakyRELU(self.corr(F, c36, warp6v))
		x = F.concat(corr6u, corr6v, flow6, dim=1)
		x = F.concat(self.conv6_0(x), x, dim=1)
		x = F.concat(self.conv6_1(x), x, dim=1)
		x = F.concat(self.conv6_2(x), x, dim=1)
		x = F.concat(self.conv6_3(x), x, dim=1)
		x = F.concat(self.conv6_4(x), x, dim=1)
		flow6 = flow6 + self.pred_flow6(x)

		feat5 = self.upfeat5(x)
		flow5 = Upsample(2)(flow6)
		warp5u = self.deform5(c25, F.repeat(F.expand_dims(flow5*self.scale/self.strides[1], axis=1), 9, axis=1).reshape((0, -3, -2)))
		warp5u = self.leakyRELU(warp5u)
		corr5u = self.leakyRELU(self.corr(F, c15, warp5u))
		warp5v = c45
		corr5v = self.leakyRELU(self.corr(F, c35, warp5v))
		x = F.concat(c15, feat5, corr5u, corr5v, flow5, flows[1], dim=1)
		x = F.concat(self.conv5_0(x), x, dim=1)
		x = F.concat(self.conv5_1(x), x, dim=1)
		x = F.concat(self.conv5_2(x), x, dim=1)
		x = F.concat(self.conv5_3(x), x, dim=1)
		x = F.concat(self.conv5_4(x), x, dim=1)
		flow5 = flow5 + self.pred_flow5(x)

		feat4 = self.upfeat4(x)
		flow4 = Upsample(2)(flow5)
		warp4u = self.deform4(c24, F.repeat(F.expand_dims(flow4*self.scale/self.strides[2], axis=1), 9, axis=1).reshape((0, -3, -2)))
		warp4u = self.leakyRELU(warp4u)
		corr4u = self.leakyRELU(self.corr(F, c14, warp4u))
		warp4v = c44
		corr4v = self.leakyRELU(self.corr(F, c34, warp4v))
		x = F.concat(c14, feat4, corr4u, corr4v, flow4, flows[2], dim=1)
		x = F.concat(self.conv4_0(x), x, dim=1)
		x = F.concat(self.conv4_1(x), x, dim=1)
		x = F.concat(self.conv4_2(x), x, dim=1)
		x = F.concat(self.conv4_3(x), x, dim=1)
		x = F.concat(self.conv4_4(x), x, dim=1)
		flow4 = flow4 + self.pred_flow4(x)

		feat3 = self.upfeat3(x)
		flow3 = Upsample(2)(flow4)
		warp3u = self.deform3(c23, F.repeat(F.expand_dims(flow3*self.scale/self.strides[3], axis=1), 9, axis=1).reshape((0, -3, -2)))
		warp3u = self.leakyRELU(warp3u)
		corr3u = self.leakyRELU(self.corr(F, c13, warp3u))
		warp3v = c43
		corr3v = self.leakyRELU(self.corr(F, c33, warp3v))
		x = F.concat(c13, feat3, corr3u, corr3v, flow3, flows[3], dim=1)
		x = F.concat(self.conv3_0(x), x, dim=1)
		x = F.concat(self.conv3_1(x), x, dim=1)
		x = F.concat(self.conv3_2(x), x, dim=1)
		x = F.concat(self.conv3_3(x), x, dim=1)
		x = F.concat(self.conv3_4(x), x, dim=1)
		flow3 = flow3 + self.pred_flow3(x)

		feat2 = self.upfeat2(x)
		flow2 = Upsample(2)(flow3)
		warp2u = self.deform2(c22, F.repeat(F.expand_dims(flow2*self.scale/self.strides[4], axis=1), 9, axis=1).reshape((0, -3, -2)))
		warp2u = self.leakyRELU(warp2u)
		corr2u = self.leakyRELU(self.corr(F, c12, warp2u))
		warp2v = c42
		corr2v = self.leakyRELU(self.corr(F, c32, warp2v))
		x = F.concat(c12, feat2, corr2u, corr2v, flow2, flows[4], dim=1)
		x = F.concat(self.conv2_0(x), x, dim=1)
		x = F.concat(self.conv2_1(x), x, dim=1)
		x = F.concat(self.conv2_2(x), x, dim=1)
		x = F.concat(self.conv2_3(x), x, dim=1)
		x = F.concat(self.conv2_4(x), x, dim=1)
		flow2 = flow2 + self.pred_flow2(x)


		x = self.dc_conv4(self.dc_conv3(self.dc_conv2(self.dc_conv1(x))))
		flow2 = flow2 + self.dc_conv7(self.dc_conv6(self.dc_conv5(x)))

		preds = [flow * self.scale for flow in [flow6, flow5, flow4, flow3, flow2]]
		visuals = []
		visuals.append(flow2.slice_axis(axis = 1, begin = 0, end = 1))
		return preds, visuals, []


class EpeLoss(nn.HybridBlock):
	''' Compute Endpoint Error Loss
	Arguments
	==============
	- pred [N, C, H, W] : predictions
	- label [N, C, H, W] : flow_groundtruth
	'''
	def __init__(self, eps = 0, **kwargs):
		super().__init__(**kwargs)
		self.eps = eps

	def hybrid_forward(self, F, pred, label):
		loss = F.sqrt(F.sum(F.square(pred - label), axis=1) + self.eps)
		return F.mean(loss, axis=0, exclude=True)

class EpeLossWithMask(nn.HybridBlock):
	''' Compute Endpoint Error Loss
	Arguments
	==============
	- pred [N, C, H, W] : predictions
	- label [N, C, H, W] : flow_groundtruth
	- mask [N, 1, H, W] : mask_groundtruth
	'''
	def __init__(self, eps = 1e-8, q = None, **kwargs):
		super().__init__(**kwargs)
		self.eps = eps
		self.q = q

	def hybrid_forward(self, F, pred, label, mask):
		if self.q is not None:
			loss = (F.sum(F.abs(pred - label), axis = 1) + self.eps) ** self.q
		else:
			loss = F.sqrt(F.sum(F.square(pred - label), axis = 1) + self.eps)
		loss = F.broadcast_mul(loss, mask.squeeze(axis = 1))
		loss = F.sum(loss, axis=0, exclude=True) / F.sum(mask, axis=0, exclude=True)
		return loss

class MultiscaleEpe(nn.HybridBlock):
	def __init__(self, scales, weights, match, eps = 1e-8, q = None, **kwargs):
		super().__init__(**kwargs)

		self.scales = scales
		self.weights = weights
		self.match = match
		self.eps = eps
		self.q = q
		if match == 'upsampling':
			with self.name_scope():
				for s in self.scales:
					setattr(self, 'upsampler_{}'.format(s), Upsample(s))

	def _get_upsampler(self, s):
		return getattr(self, 'upsampler_{}'.format(s))

	def hybrid_forward(self, F, flow, mask, *predictions):
		if self.match == 'upsampling':
			losses = [EpeLossWithMask(eps = self.eps, q = self.q)(self._get_upsampler(s)(p), flow, mask) * w
				for p, w, s in zip(predictions, self.weights, self.scales)]
		elif self.match == 'downsampling':
			losses = [EpeLossWithMask(eps = self.eps, q = self.q)(p, Downsample(s)(flow), Downsample(s)(mask)) * w
				for p, w, s in zip(predictions, self.weights, self.scales)]
		else:
			raise NotImplementedError
		return F.add_n(*losses)



# SPDX-License-Identifier: MIT
# Source: https://github.com/microsoft/MaskFlownet/tree/5cba12772e2201f0d1c1e27161d224e585334571
import os
import sys

# ======== PLEASE MODIFY ========
# where is the repo
repoRoot = r'.'
# to CUDA\vX.Y\bin
#os.environ['PATH'] = r'path\to\your\NVIDIA GPU Computing Toolkit\CUDA\v9.0\bin' + ';' + os.environ['PATH']

import argparse
import yaml
import numpy as np
import mxnet as mx
import cv2
import flow_vis
from moviepy.editor import ImageSequenceClip
from moviepy.audio.AudioClip import AudioArrayClip

from .network import config
from .network import get_pipeline
from . import path
from . import logger


def find_checkpoint(checkpoint_str):
    # find checkpoint
    steps = 0
    if checkpoint_str is not None:
    	if ':' in checkpoint_str:
    		prefix, steps = checkpoint_str.split(':')
    	else:
    		prefix = checkpoint_str
    		steps = None
    	log_file, run_id = path.find_log(prefix)	
    	if steps is None:
    		checkpoint, steps = path.find_checkpoints(run_id)[-1]
    	else:
    		checkpoints = path.find_checkpoints(run_id)
    		try:
    			checkpoint, steps = next(filter(lambda t : t[1] == steps, checkpoints))
    		except StopIteration:
    			print('The steps not found in checkpoints', steps, checkpoints)
    			sys.stdout.flush()
    			raise StopIteration
    	steps = int(steps)
    	if args.clear_steps:
    		steps = 0
    	else:
    		_, exp_info = path.read_log(log_file)
    		exp_info = exp_info[-1]
    		for k in args.__dict__:
    			if k in exp_info and k in ('tag',):
    				setattr(args, k, eval(exp_info[k]))
    				print('{}={}, '.format(k, exp_info[k]), end='')
    		print()
    	sys.stdout.flush()
    return checkpoint, steps


def load_model(config_str):
    # load configuration
    with open(os.path.join(repoRoot, 'network', 'config', config_str)) as f:
    	config =  config.Reader(yaml.load(f))
    return config


def instantiate_model(gpu_device, config):
    ctx = [mx.cpu()] if gpu_device == '' else [mx.gpu(gpu_id) for gpu_id in map(int, gpu_device.split(','))]
    # initiate
    pipe = get_pipeline(args.network, ctx=ctx, config=config)
    return pipe
    

def load_checkpoint(pipe, config, checkpoint):
    # load parameters from given checkpoint
    print('Load Checkpoint {}'.format(checkpoint))
    sys.stdout.flush()
    network_class = getattr(config.network, 'class').get()
    print('load the weight for the network')
    pipe.load(checkpoint)
    if network_class == 'MaskFlownet':
   		print('fix the weight for the head network')
   		pipe.fix_head()
    sys.stdout.flush()
    return pipe


def predict_image_pair_flow(img1, img2, pipe, resize=None):
    for result in pipe.predict([img1], [img2], batch_size = 1, resize=resize):
        flow, occ_mask, warped = result
    return flow, occ_mask, warped
    

def create_video_clip_from_frames(frame_list, fps):
    """ Function takes a list of video frames and puts them together in a sequence"""
    visual_clip = ImageSequenceClip(frame_list, fps=fps) #put frames together using moviepy
    return visual_clip #return the ImageSequenceClip


def predict_video_flow(video_filename, batch_size, resize=None):
    cap = cv2.VideoCapture(video_filename)
    fps = cap.get(cv2.CAP_PROP_FPS)

    prev_frames = []
    new_frames = []
    has_frames, frame = cap.read()
    prev_frames.append(frame)
    while True:
        has_frames, frame = cap.read()
        if not has_frames:
            cap.release()
            break
        new_frames.append(frame)
        prev_frames.append(frame)
    del prev_frames[-1] #delete the last frame of the video from prev_frames
    flow_video = [flow for flow, occ_mask, warped in pipe.predict(prev_frames, new_frames, batch_size=batch_size, resize=resize)]
    
    return flow_video, fps



if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument('flow_filepath', type=str, help='destination filepath of the flow image/video')
    parser.add_argument('config', type=str, nargs='?', default=None)
    parser.add_argument('--image_1', type=str, help='filepath of the first image')
    parser.add_argument('--image_2', type=str, help='filepath of the second image')
    parser.add_argument('--video_filepath', type=str, help='filepath of the input video')
    parser.add_argument('-g', '--gpu_device', type=str, default='', help='Specify gpu device(s)')
    parser.add_argument('-c', '--checkpoint', type=str, default=None, 
    	help='model checkpoint to load; by default, the latest one.'
    	'You can use checkpoint:steps to load to a specific steps')
    parser.add_argument('--clear_steps', action='store_true')
    parser.add_argument('-n', '--network', type=str, default='MaskFlownet', help='The choice of network')
    parser.add_argument('--batch', type=int, default=8, help='minibatch size of samples per device')
    parser.add_argument('--resize', type=str, default='', help='shape to resize image frames before inference')
    parser.add_argument('--threads', type=str, default=8, help='Number of threads to use when writing flow video to file')
    
    args = parser.parse_args()
    
    
    # Get desired image resize from the string argument
    infer_resize = [int(s) for s in args.resize.split(',')] if args.resize else None
    
    checkpoint, steps = find_checkpoint(args.checkpoint)
    config = load_model(args.config)
    pipe = instantiate_model(args.gpu_device, config)
    pipe = load_checkpoint(pipe, config, checkpoint)
    
    if args.image_1 is not None:
        image_1 = cv2.imread(args.image_1)
        image_2 = cv2.imread(args.image_2)
        flow, occ_mask, warped = predict_image_pair_flow(image_1, image_2, pipe)
        cv2.imwrite(args.flow_filepath, flow_vis.flow_to_color(flow, convert_to_bgr=False))
    else:
        flow_video, fps = predict_video_flow(args.video_filepath, batch_size=args.batch)
        flow_video_visualisations = [flow_vis.flow_to_color(flow, convert_to_bgr=False) for flow in flow_video]
        flow_video_clip = create_video_clip_from_frames(flow_video_visualisations, fps)
        flow_video_clip.write_videofile(args.flow_filepath, threads=args.threads, logger=None) #export the video

    sys.exit(0)



# Copyright 2021, Blue Brain Project, EPFL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for optical models."""
import argparse
from unittest import mock

import numpy as np
import pytest
import torch

from atlinter.data import GeneDataset
from atlinter.optical_flow import GeneOpticalFlow, MaskFlowNet, OpticalFlow, RAFTNet


class DummyModel(OpticalFlow):
    def predict_flow(self, img1, img2):
        raise NotImplementedError


@pytest.mark.parametrize("rgb", [True, False])
def test_optical_flow_model(rgb):
    """Test Optical Flow."""
    if rgb:
        shape = (10, 10, 3)
    else:
        shape = (10, 10)

    img1 = np.random.rand(*shape)
    img2 = np.random.rand(*shape)
    flow = np.random.rand(2, 10, 10)

    model = DummyModel()

    pre_img1, pre_img2 = model.preprocess_images(img1, img2)
    warped_img2 = model.warp_image(flow, img2)
    assert pre_img1 is img1
    assert pre_img2 is img2
    assert warped_img2.shape == shape

    with pytest.raises(ValueError) as exc:
        img2 = np.ones((20, 20, 3))
        model.warp_image(flow, img2)
    assert "The flow shape and the image shape should be consistent." in str(exc.value)

    with pytest.raises(ValueError) as exc:
        img2 = np.ones((10, 10, 1, 1, 5))
        model.warp_image(flow, img2)
    assert "Invalid image shape:" in str(exc.value)


@pytest.mark.parametrize("rgb", [True, False])
def test_raftnet_model(monkeypatch, rgb):
    """Test RAFT model."""
    if rgb:
        shape = (10, 10, 3)
    else:
        shape = (10, 10)
    img1 = np.random.rand(*shape)
    img2 = np.random.rand(*shape)

    if rgb:
        img1 = (img1 * 255).astype(np.uint8)
        img2 = (img2 * 255).astype(np.uint8)

    fake_load = mock.Mock(return_value={})
    fake_torch_parallel = mock.Mock()
    fake_load_state = mock.Mock()
    fake_torch_parallel.load_state_dict.return_value = fake_load_state
    monkeypatch.setattr(
        "atlinter.optical_flow.torch.nn.DataParallel", fake_torch_parallel
    )
    monkeypatch.setattr("atlinter.optical_flow.torch.load", fake_load)

    raft = RAFTNet(path="test", device="cpu")
    raft.model = mock.Mock(
        spec=torch.nn.Module,
        return_value=[torch.rand(1, 2, 10, 10), torch.rand(1, 2, 10, 10)],
    )

    # ValueError when shape of images are different.
    with pytest.raises(ValueError):
        raft.preprocess_images(img1, np.random.rand(20, 20))

    args = raft.initialize_namespace()
    assert isinstance(args, argparse.Namespace)

    preimg1, preimg2 = raft.preprocess_images(img1, img2)
    assert isinstance(preimg1, np.ndarray) and isinstance(preimg2, np.ndarray)
    expected_shape = (1, 3, 16, 16)
    assert (preimg1.shape == expected_shape) and (preimg2.shape == expected_shape)
    assert preimg1.dtype == np.uint8 and preimg2.dtype == np.uint8
    assert 1 < np.max(preimg1) <= 255 and 1 < np.max(preimg2) <= 255

    flow = raft.predict_flow(preimg1, preimg2)
    assert isinstance(flow, np.ndarray)
    assert flow.shape == (2, 10, 10)

    new_img1 = raft.warp_image(flow, img2)
    assert isinstance(new_img1, np.ndarray)
    assert new_img1.shape == shape


@pytest.mark.parametrize("rgb", [True, False])
def test_maskflownet_model(monkeypatch, rgb):
    """Test RAFT model."""
    if rgb:
        shape = (10, 10, 3)
    else:
        shape = (10, 10)
    img1 = np.random.rand(*shape)
    img2 = np.random.rand(*shape)

    if rgb:
        img1 = (img1 * 255).astype(np.uint8)
        img2 = (img2 * 255).astype(np.uint8)

    fake_network = mock.Mock()
    fake_network.get_pipeline.return_value = "pipe"
    monkeypatch.setattr("atlinter.optical_flow.MaskFlowNet.network", fake_network)
    fake_predict_new_data = mock.Mock()
    fake_predict_new_data.load_checkpoint.return_value = "model"
    fake_predict_new_data.predict_image_pair_flow.return_value = (
        np.random.rand(10, 10, 2),
        img1,
        img1,
    )
    monkeypatch.setattr(
        "atlinter.optical_flow.MaskFlowNet.predict_new_data", fake_predict_new_data
    )

    maskflow = MaskFlowNet(checkpoint_path="fake")
    with pytest.raises(ValueError):
        maskflow.preprocess_images(img1, np.random.rand(20, 20))

    preimg1, preimg2 = maskflow.preprocess_images(img1, img2)
    assert isinstance(preimg1, np.ndarray) and isinstance(preimg2, np.ndarray)
    expected_shape = (10, 10, 3)
    assert (preimg1.shape == expected_shape) and (preimg2.shape == expected_shape)
    assert preimg1.dtype == np.uint8 and preimg2.dtype == np.uint8
    assert 1 < np.max(preimg1) <= 255 and 1 < np.max(preimg2) <= 255

    flow = maskflow.predict_flow(preimg1, preimg2)
    assert isinstance(flow, np.ndarray)
    assert flow.shape == (2, 10, 10)

    new_img1 = maskflow.warp_image(flow, img2)
    assert isinstance(new_img1, np.ndarray)
    assert new_img1.shape == shape


class FakeModel(OpticalFlow):
    def predict_flow(self, img1, img2):
        return np.random.randn(2, *img1.shape[:2])


def instantiate_gene_optical_flow(gene_shape, volume_shape, axis):

    gene_data = GeneDataset(
        np.ones((3, *gene_shape)),
        [3, 8, 12],  # random
        axis=axis,
        volume_shape=volume_shape,
    )
    fake_model = FakeModel()
    reference_space = np.random.randn(*volume_shape)
    return GeneOpticalFlow(gene_data, reference_space, fake_model)


class TestGeneOpticalFlow:
    @pytest.mark.parametrize("axis", ["coronal", "sagittal"])
    @pytest.mark.parametrize("rgb", [True, False])
    @pytest.mark.parametrize("volume_shape", [(20, 30, 40), (40, 10, 35)])
    def test_predict_ref_flow(self, axis, rgb, volume_shape):
        """Test get prediction."""
        if axis == "coronal":
            gene_shape = (volume_shape[1], volume_shape[2])
        else:
            gene_shape = (volume_shape[0], volume_shape[1])

        if rgb:
            volume_shape = (*volume_shape, 3)
            gene_shape = (*gene_shape, 3)

        gene_optical_flow = instantiate_gene_optical_flow(
            gene_shape, volume_shape, axis
        )
        assert gene_optical_flow.axis == axis
        assert gene_optical_flow.pairwise_ref_flows == {}

        prediction_1 = gene_optical_flow.predict_ref_flow(0, 3)
        assert isinstance(prediction_1, np.ndarray)
        assert prediction_1.shape == (2, *gene_shape[0:2])
        assert (0, 3) in gene_optical_flow.pairwise_ref_flows
        prediction_2 = gene_optical_flow.predict_ref_flow(0, 3)
        assert np.all(prediction_1 == prediction_2)

        with pytest.raises(ValueError):
            # Out of boundary of the reference space
            gene_optical_flow.predict_ref_flow(0, 100)

    @pytest.mark.parametrize("axis", ["coronal", "sagittal"])
    @pytest.mark.parametrize("rgb", [True, False])
    @pytest.mark.parametrize("volume_shape", [(20, 30, 40), (40, 10, 35)])
    def test_predict_slice(self, axis, rgb, volume_shape):
        """Test prediction of slice"""
        if axis == "coronal":
            gene_shape = (volume_shape[1], volume_shape[2])
        else:
            gene_shape = (volume_shape[0], volume_shape[1])

        if rgb:
            volume_shape = (*volume_shape, 3)
            gene_shape = (*gene_shape, 3)

        gene_optical_flow = instantiate_gene_optical_flow(
            gene_shape, volume_shape, axis
        )
        predicted_slice = gene_optical_flow.predict_slice(10)
        assert isinstance(predicted_slice, np.ndarray)
        assert predicted_slice.shape == gene_shape

    @pytest.mark.parametrize("axis", ["coronal", "sagittal"])
    @pytest.mark.parametrize("rgb", [True, False])
    @pytest.mark.parametrize("volume_shape", [(20, 30, 40), (40, 10, 35)])
    def test_predict_volume(self, axis, rgb, volume_shape):
        """Test prediction of the volume"""
        if axis == "coronal":
            gene_shape = (volume_shape[1], volume_shape[2])
        else:
            gene_shape = (volume_shape[0], volume_shape[1])

        if rgb:
            volume_shape = (*volume_shape, 3)
            gene_shape = (*gene_shape, 3)

        gene_optical_flow = instantiate_gene_optical_flow(
            gene_shape, volume_shape, axis
        )
        predicted_volume = gene_optical_flow.predict_volume()
        assert isinstance(predicted_volume, np.ndarray)
        assert predicted_volume.shape == volume_shape
        # Check that the known slices are kept in the volume
        if axis == "sagittal":
            predicted_volume = np.moveaxis(predicted_volume, 2, 0)
        assert np.all(predicted_volume[[3, 8, 12]] == np.ones((3, *gene_shape)))



# SPDX-License-Identifier: MIT
# Source: https://github.com/myungsub/CAIN/tree/2e727d2a07d3f1061f17e2edaa47a7fb3f7e62c5
import math
import numpy as np

import torch
import torch.nn as nn

from .common import *

class Encoder(nn.Module):
    def __init__(self, in_channels=3, depth=3):
        super(Encoder, self).__init__()
        self.device = torch.device('cuda')
        
        self.shuffler = PixelShuffle(1/2**depth)
        # self.shuffler = nn.Sequential(
        #    PixelShuffle(1/2),
        #    PixelShuffle(1/2),
        #    PixelShuffle(1/2))
        self.interpolate = Interpolation_res(5, 12, in_channels * (4**depth))

    def forward(self, x1, x2):
        feats1 = self.shuffler(x1)
        feats2 = self.shuffler(x2)

        feats = self.interpolate(feats1, feats2)

        return feats


class Decoder(nn.Module):
    def __init__(self, depth=3):
        super(Decoder, self).__init__()
        self.device = torch.device('cuda')

        self.shuffler = PixelShuffle(2**depth)
        # self.shuffler = nn.Sequential(
        #    PixelShuffle(2),
        #    PixelShuffle(2),
        #    PixelShuffle(2))

    def forward(self, feats):
        out = self.shuffler(feats)
        return out


class CAIN_NoCA(nn.Module):
    def __init__(self, depth=3):
        super(CAIN_NoCA, self).__init__()
        self.depth = depth

        self.encoder = Encoder(in_channels=3, depth=depth)
        self.decoder = Decoder(depth=depth)

    def forward(self, x1, x2):
        x1, m1 = sub_mean(x1)
        x2, m2 = sub_mean(x2)

        if not self.training:
            paddingInput, paddingOutput = InOutPaddings(x1)
            x1 = paddingInput(x1)
            x2 = paddingInput(x2)

        feats = self.encoder(x1, x2)
        out = self.decoder(feats)

        if not self.training:
            out = paddingOutput(out)

        mi = (m1 + m2) / 2
        out += mi

        return out, feats



# Copyright 2021, Blue Brain Project, EPFL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Align special volumes to Nissl slices."""
import argparse
import pathlib
import sys

import numpy as np
from atlannot.ants import register, transform


def parse_args():
    """Parse command line arguments.

    Returns
    -------
    args : argparse.Namespace
        The parsed command line arguments.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("dataset_id")
    args = parser.parse_args()

    return args


def dataset_registration(nissl, gene):
    """Align gene dataset to Nissl.

    Parameters
    ----------
    nissl : np.ndarray
        Corresponding nissl section for the given gene.
    gene : np.ndarray
        Dataset of gene expression images to align to Nissl.

    Returns
    -------
    aligned_gene : np.ndarray
        Aligned gene expressions.
    """
    aligned_gene = []
    for section_number, single_gene in enumerate(gene):
        if np.sum(single_gene) == 0:
            aligned_gene.append(single_gene)
            continue
        else:
            try:
                single_nissl = nissl[int(section_number), :, :]
            except IndexError:
                aligned_gene += [
                    single_gene,
                ]
                continue

        aligned_gene += [
            single_registration(single_nissl, single_gene),
        ]

    aligned_gene = np.array(aligned_gene)
    return aligned_gene


def single_registration(single_nissl, single_gene):
    """Align ISH data to a given nissl slice thanks to ANTsPY.

    Parameters
    ----------
    single_nissl : np.ndarray
        Corresponding nissl section for the given gene.
    single_gene : np.ndarray
        Gene expression to align to Nissl.

    Returns
    -------
    aligned_single_gene : np.ndarray
        Aligned gene expression.
    """
    tmp_gene = single_gene.copy()

    if single_nissl.shape != tmp_gene.shape:
        raise ValueError(
            f"Nissl ({single_nissl.shape}) and gene "
            f"({tmp_gene.shape}) should have the same shape"
        )

    if np.max(single_gene) > 1:
        tmp_gene = single_gene / single_gene.max()

    nii_data = register(single_nissl, tmp_gene.astype(np.float32))
    aligned_single_gene = transform(single_gene.astype(np.float32), nii_data)

    return aligned_single_gene


def main():
    """Align gene expression to Nissl."""
    # Imports
    import nrrd

    args = parse_args()
    dataset_id = args.dataset_id

    cwd = pathlib.Path(__file__).parent.resolve()
    nissl_path = cwd / "ara_nissl_25.nrrd"
    if not nissl_path.exists():
        raise ValueError("Nissl volume is missing! Please dvc pull download-nissl")

    nissl, _ = nrrd.read(nissl_path)
    nissl = nissl / nissl.max()

    file_dir = cwd / "special_volumes" / f"dataset_{dataset_id}"
    if not file_dir.exists():
        raise ValueError(f"The dataset {dataset_id} does not exist.")

    experiment_ids = [
        path.stem
        for path in file_dir.iterdir()
        if path.suffix == ".npy" and path.stem != "dfs"
    ]

    saving_dir = cwd / "aligned" / "special_volumes" / f"dataset_{dataset_id}"
    if not saving_dir.exists():
        saving_dir.mkdir(parents=True)

    for experiment in experiment_ids:
        # Loading
        gene = np.load(file_dir / f"{experiment}.npy")

        # Registration
        aligned_gene = dataset_registration(nissl, gene)

        # Saving
        np.save(saving_dir / f"{experiment}.npy", aligned_gene)


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2021, Blue Brain Project, EPFL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implementation of the Fréchet Inception Distance (FID) metric."""
import numpy as np
import torch
from pytorch_fid.fid_score import calculate_frechet_distance
from pytorch_fid.inception import InceptionV3


class FID:
    """Class for computing the Fréchet Inception Distance (FID) metric score.

    It is implemented as a class in order to hold the inception model instance
    in its state.

    Parameters
    ----------
    resize_input : bool (optional)
        Whether or not to resize the input images to the image size (299, 299)
        on which the inception model was trained. Since the model is fully
        convolutional, the score also works without resizing. In literature
        and when working with GANs people tend to set this value to True,
        however, for internal evaluation this is not necessary.
    device : str or torch.device
        The device on which to run the inception model.
    """

    def __init__(self, resize_input=False, device=None):
        self.device = device
        if self.device is None:
            self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.model = InceptionV3(resize_input=resize_input).to(device)
        self.model = self.model.eval()

    def _get_mu_sigma(self, images):
        """Compute the inception mu and sigma for a batch of images.

        Parameters
        ----------
        images : np.ndarray
            A batch of images with shape (n_images, width, height).

        Returns
        -------
        mu : np.ndarray
            The array of mean activations with shape (2048,).
        sigma : np.ndarray
            The covariance matrix of activations with shape (2048, 2048).
        """
        # forward pass
        batch = torch.tensor(images).unsqueeze(1).repeat((1, 3, 1, 1))
        batch = batch.to(self.device, torch.float32)
        (activations,) = self.model(batch)
        activations = activations.detach().cpu().numpy().squeeze(3).squeeze(2)

        # compute statistics
        mu = activations.mean(axis=0)
        sigma = np.cov(activations, rowvar=False)

        return mu, sigma

    def score(self, images_1, images_2):
        """Compute the FID score.

        The input batches should have the shape (n_images, width, height).

        Parameters
        ----------
        images_1 : np.ndarray
            First batch of images.
        images_2 : np.ndarray
            Section batch of images.

        Returns
        -------
        score : float
            The FID score.
        """
        mu_1, sigma_1 = self._get_mu_sigma(images_1)
        mu_2, sigma_2 = self._get_mu_sigma(images_2)
        score = calculate_frechet_distance(mu_1, sigma_1, mu_2, sigma_2)

        return score



# Copyright 2021, Blue Brain Project, EPFL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
import pytest

from atlinter.utils import find_closest


@pytest.mark.parametrize(
    ("x", "values", "expected_index", "expected_value"),
    [
        (6.9, np.arange(10), 7, 7),
        (3, np.arange(10), 3, 3),
        (3, np.arange(1, 10), 2, 3),
        (9.4, [0.1, 4.3, 9.5], 2, 9.5),
    ],
)
def test_find_closest(x, values, expected_index, expected_value):
    """Test find closest function."""
    assert find_closest(x, values) == (expected_index, expected_value)



# SPDX-License-Identifier: MIT
# Source: https://github.com/myungsub/CAIN/tree/2e727d2a07d3f1061f17e2edaa47a7fb3f7e62c5
import math

import torch
import torch.nn as nn
import torch.nn.functional as F

def sub_mean(x):
    mean = x.mean(2, keepdim=True).mean(3, keepdim=True)
    x -= mean
    return x, mean

def InOutPaddings(x):
    w, h = x.size(3), x.size(2)
    padding_width, padding_height = 0, 0
    if w != ((w >> 7) << 7):
        padding_width = (((w >> 7) + 1) << 7) - w
    if h != ((h >> 7) << 7):
        padding_height = (((h >> 7) + 1) << 7) - h
    paddingInput = nn.ReflectionPad2d(padding=[padding_width // 2, padding_width - padding_width // 2,
                                               padding_height // 2, padding_height - padding_height // 2])
    paddingOutput = nn.ReflectionPad2d(padding=[0 - padding_width // 2, padding_width // 2 - padding_width,
                                                0 - padding_height // 2, padding_height // 2 - padding_height])
    return paddingInput, paddingOutput


class ConvNorm(nn.Module):
    def __init__(self, in_feat, out_feat, kernel_size, stride=1, norm=False):
        super(ConvNorm, self).__init__()

        reflection_padding = kernel_size // 2
        self.reflection_pad = nn.ReflectionPad2d(reflection_padding)
        self.conv = nn.Conv2d(in_feat, out_feat, stride=stride, kernel_size=kernel_size, bias=True)

        self.norm = norm
        if norm == 'IN':
            self.norm = nn.InstanceNorm2d(out_feat, track_running_stats=True)
        elif norm == 'BN':
            self.norm = nn.BatchNorm2d(out_feat)

    def forward(self, x):
        out = self.reflection_pad(x)
        out = self.conv(out)
        if self.norm:
            out = self.norm(out)
        return out


class UpConvNorm(nn.Module):
    def __init__(self, in_channels, out_channels, mode='transpose', norm=False):
        super(UpConvNorm, self).__init__()

        if mode == 'transpose':
            self.upconv = nn.ConvTranspose2d(in_channels, out_channels, kernel_size=4, stride=2, padding=1)
        elif mode == 'shuffle':
            self.upconv = nn.Sequential(
                ConvNorm(in_channels, 4*out_channels, kernel_size=3, stride=1, norm=norm),
                PixelShuffle(2))
        else:
            # out_channels is always going to be the same as in_channels
            self.upconv = nn.Sequential(
                nn.Upsample(mode='bilinear', scale_factor=2, align_corners=False),
                ConvNorm(in_channels, out_channels, kernel_size=1, stride=1, norm=norm))
    
    def forward(self, x):
        out = self.upconv(x)
        return out



class meanShift(nn.Module):
    def __init__(self, rgbRange, rgbMean, sign, nChannel=3):
        super(meanShift, self).__init__()
        if nChannel == 1:
            l = rgbMean[0] * rgbRange * float(sign)

            self.shifter = nn.Conv2d(1, 1, kernel_size=1, stride=1, padding=0)
            self.shifter.weight.data = torch.eye(1).view(1, 1, 1, 1)
            self.shifter.bias.data = torch.Tensor([l])
        elif nChannel == 3:  
            r = rgbMean[0] * rgbRange * float(sign)
            g = rgbMean[1] * rgbRange * float(sign)
            b = rgbMean[2] * rgbRange * float(sign)

            self.shifter = nn.Conv2d(3, 3, kernel_size=1, stride=1, padding=0)
            self.shifter.weight.data = torch.eye(3).view(3, 3, 1, 1)
            self.shifter.bias.data = torch.Tensor([r, g, b])
        else:
            r = rgbMean[0] * rgbRange * float(sign)
            g = rgbMean[1] * rgbRange * float(sign)
            b = rgbMean[2] * rgbRange * float(sign)
            self.shifter = nn.Conv2d(6, 6, kernel_size=1, stride=1, padding=0)
            self.shifter.weight.data = torch.eye(6).view(6, 6, 1, 1)
            self.shifter.bias.data = torch.Tensor([r, g, b, r, g, b])

        # Freeze the meanShift layer
        for params in self.shifter.parameters():
            params.requires_grad = False

    def forward(self, x):
        x = self.shifter(x)

        return x


""" CONV - (BN) - RELU - CONV - (BN) """
class ResBlock(nn.Module):
    def __init__(self, in_feat, out_feat, kernel_size=3, reduction=False, bias=True, # 'reduction' is just for placeholder
                 norm=False, act=nn.ReLU(True), downscale=False):
        super(ResBlock, self).__init__()

        self.body = nn.Sequential(
            ConvNorm(in_feat, out_feat, kernel_size=kernel_size, stride=2 if downscale else 1),
            act,
            ConvNorm(out_feat, out_feat, kernel_size=kernel_size, stride=1)
        )
        
        self.downscale = None
        if downscale:
            self.downscale = nn.Conv2d(in_feat, out_feat, kernel_size=1, stride=2)

    def forward(self, x):
        res = x
        out = self.body(x)
        if self.downscale is not None:
            res = self.downscale(res)
        out += res

        return out 


## Channel Attention (CA) Layer
class CALayer(nn.Module):
    def __init__(self, channel, reduction=16):
        super(CALayer, self).__init__()
        # global average pooling: feature --> point
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        # feature channel downscale and upscale --> channel weight
        self.conv_du = nn.Sequential(
            nn.Conv2d(channel, channel // reduction, 1, padding=0, bias=True),
            nn.ReLU(inplace=True),
            nn.Conv2d(channel // reduction, channel, 1, padding=0, bias=True),
            nn.Sigmoid()
        )

    def forward(self, x):
        y = self.avg_pool(x)
        y = self.conv_du(y)
        return x * y, y


## Residual Channel Attention Block (RCAB)
class RCAB(nn.Module):
    def __init__(self, in_feat, out_feat, kernel_size, reduction, bias=True,
            norm=False, act=nn.ReLU(True), downscale=False, return_ca=False):
        super(RCAB, self).__init__()

        self.body = nn.Sequential(
            ConvNorm(in_feat, out_feat, kernel_size, stride=2 if downscale else 1, norm=norm),
            act,
            ConvNorm(out_feat, out_feat, kernel_size, stride=1, norm=norm),
            CALayer(out_feat, reduction)
        )
        self.downscale = downscale
        if downscale:
            self.downConv = nn.Conv2d(in_feat, out_feat, kernel_size=3, stride=2, padding=1)
        self.return_ca = return_ca

    def forward(self, x):
        res = x
        out, ca = self.body(x)
        if self.downscale:
            res = self.downConv(res)
        out += res

        if self.return_ca:
            return out, ca
        else:
            return out


## Residual Group (RG)
class ResidualGroup(nn.Module):
    def __init__(self, Block, n_resblocks, n_feat, kernel_size, reduction, act, norm=False):
        super(ResidualGroup, self).__init__()

        modules_body = [Block(n_feat, n_feat, kernel_size, reduction, bias=True, norm=norm, act=act)
            for _ in range(n_resblocks)]
        modules_body.append(ConvNorm(n_feat, n_feat, kernel_size, stride=1, norm=norm))
        self.body = nn.Sequential(*modules_body)

    def forward(self, x):
        res = self.body(x)
        res += x
        return res


def pixel_shuffle(input, scale_factor):
    batch_size, channels, in_height, in_width = input.size()

    out_channels = int(int(channels / scale_factor) / scale_factor)
    out_height = int(in_height * scale_factor)
    out_width = int(in_width * scale_factor)

    if scale_factor >= 1:
        input_view = input.contiguous().view(batch_size, out_channels, scale_factor, scale_factor, in_height, in_width)
        shuffle_out = input_view.permute(0, 1, 4, 2, 5, 3).contiguous()
    else:
        block_size = int(1 / scale_factor)
        input_view = input.contiguous().view(batch_size, channels, out_height, block_size, out_width, block_size)
        shuffle_out = input_view.permute(0, 1, 3, 5, 2, 4).contiguous()

    return shuffle_out.view(batch_size, out_channels, out_height, out_width)


class PixelShuffle(nn.Module):
    def __init__(self, scale_factor):
        super(PixelShuffle, self).__init__()
        self.scale_factor = scale_factor
    
    def forward(self, x):
        return pixel_shuffle(x, self.scale_factor)
    def extra_repr(self):
        return 'scale_factor={}'.format(self.scale_factor)


def conv(in_channels, out_channels, kernel_size, 
         stride=1, bias=True, groups=1):
    return nn.Conv2d(
        in_channels,
        out_channels,
        kernel_size=kernel_size,
        padding=kernel_size//2,
        stride=1,
        bias=bias,
        groups=groups)


def conv1x1(in_channels, out_channels, stride=1, bias=True, groups=1):
    return nn.Conv2d(
        in_channels,
        out_channels,
        kernel_size=1,
        stride=stride,
        bias=bias,
        groups=groups)

def conv3x3(in_channels, out_channels, stride=1, 
            padding=1, bias=True, groups=1):    
    return nn.Conv2d(
        in_channels,
        out_channels,
        kernel_size=3,
        stride=stride,
        padding=padding,
        bias=bias,
        groups=groups)

def conv5x5(in_channels, out_channels, stride=1, 
            padding=2, bias=True, groups=1):    
    return nn.Conv2d(
        in_channels,
        out_channels,
        kernel_size=5,
        stride=stride,
        padding=padding,
        bias=bias,
        groups=groups)

def conv7x7(in_channels, out_channels, stride=1, 
            padding=3, bias=True, groups=1):    
    return nn.Conv2d(
        in_channels,
        out_channels,
        kernel_size=7,
        stride=stride,
        padding=padding,
        bias=bias,
        groups=groups)

def upconv2x2(in_channels, out_channels, mode='shuffle'):
    if mode == 'transpose':
        return nn.ConvTranspose2d(
            in_channels,
            out_channels,
            kernel_size=4,
            stride=2,
            padding=1)
    elif mode == 'shuffle':
        return nn.Sequential(
            conv3x3(in_channels, 4*out_channels),
            PixelShuffle(2))
    else:
        # out_channels is always going to be the same as in_channels
        return nn.Sequential(
            nn.Upsample(mode='bilinear', scale_factor=2, align_corners=False),
            conv1x1(in_channels, out_channels))



class Interpolation(nn.Module):
    def __init__(self, n_resgroups, n_resblocks, n_feats, 
                 reduction=16, act=nn.LeakyReLU(0.2, True), norm=False):
        super(Interpolation, self).__init__()

        # define modules: head, body, tail
        self.headConv = conv3x3(n_feats * 2, n_feats)

        modules_body = [
            ResidualGroup(
                RCAB,
                n_resblocks=n_resblocks,
                n_feat=n_feats,
                kernel_size=3,
                reduction=reduction, 
                act=act, 
                norm=norm)
            for _ in range(n_resgroups)]
        self.body = nn.Sequential(*modules_body)

        self.tailConv = conv3x3(n_feats, n_feats)

    def forward(self, x0, x1):
        # Build input tensor
        x = torch.cat([x0, x1], dim=1)
        x = self.headConv(x)

        res = self.body(x)
        res += x

        out = self.tailConv(res)
        return out


class Interpolation_res(nn.Module):
    def __init__(self, n_resgroups, n_resblocks, n_feats,
                 act=nn.LeakyReLU(0.2, True), norm=False):
        super(Interpolation_res, self).__init__()

        # define modules: head, body, tail (reduces concatenated inputs to n_feat)
        self.headConv = conv3x3(n_feats * 2, n_feats)

        modules_body = [ResidualGroup(ResBlock, n_resblocks=n_resblocks, n_feat=n_feats, kernel_size=3,
                            reduction=0, act=act, norm=norm)
                        for _ in range(n_resgroups)]
        self.body = nn.Sequential(*modules_body)

        self.tailConv = conv3x3(n_feats, n_feats)

    def forward(self, x0, x1):
        # Build input tensor
        x = torch.cat([x0, x1], dim=1)
        x = self.headConv(x)

        res = x
        for m in self.body:
            res = m(res)
        res += x

        x = self.tailConv(res)

        return x



# SPDX-License-Identifier: MIT
# Source: https://github.com/microsoft/MaskFlownet/tree/5cba12772e2201f0d1c1e27161d224e585334571
class Reader:
    def __init__(self, obj, full_attr=""):
        self._object = obj
        self._full_attr = full_attr
    
    def __getattr__(self, name):
        if self._object is None:
            ret = None
        else:
            ret = self._object.get(name, None)
        return Reader(ret, self._full_attr + '.' + name)

    def get(self, default=None):
        if self._object is None:
            print('Default FLAGS.{} to {}'.format(self._full_attr, default))
            return default
        else:
            return self._object

    @property
    def value(self):
        return self._object


# Copyright 2021, Blue Brain Project, EPFL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
import pytest

from atlinter.data import GeneDataset


class TestGeneData:
    @pytest.mark.parametrize("axis", ["coronal", "sagittal"])
    @pytest.mark.parametrize("rgb", [True, False])
    def test_gene_data(self, axis, rgb):
        """Test GeneData class."""
        if rgb:
            volume_shape = (10, 20, 30, 3)
        else:
            volume_shape = (10, 20, 30)
        end = volume_shape[0] if axis == "coronal" else volume_shape[2]
        gene_shape = (20, 30) if axis == "coronal" else (10, 20)
        if rgb:
            known_gene = np.random.rand(2, *gene_shape, 3)
        else:
            known_gene = np.random.rand(2, *gene_shape)
        known_slices = [5, 8]  # random

        gene_data = GeneDataset(
            known_gene,
            section_numbers=known_slices,
            axis=axis,
            volume_shape=volume_shape,
        )

        assert isinstance(gene_data.axis, str)
        assert gene_data.axis == axis
        assert gene_data.known_slices == sorted(known_slices)
        assert isinstance(gene_data.volume, np.ndarray)
        assert gene_data.volume.shape == volume_shape
        assert isinstance(gene_data.unknown_slices, list)
        assert len(gene_data.unknown_slices) == end - len(known_slices)

        # Check volume is well populated with known slices
        if gene_data.axis == "sagittal" and rgb:
            volume = np.transpose(gene_data.volume, (2, 0, 1, 3))
        elif gene_data.axis == "sagittal" and not rgb:
            volume = np.transpose(gene_data.volume, (2, 0, 1))
        else:
            volume = gene_data.volume
        for i in range(end):
            if i in known_slices:
                assert volume[i].sum() > 0
            else:
                assert volume[i].sum() == 0

    @pytest.mark.parametrize("rgb", [True, False])
    def test_get_closest(self, rgb):
        """Test get closest known from GeneData."""
        if rgb:
            image_shape = (20, 30, 3)
        else:
            image_shape = (20, 30)

        known_gene = np.random.rand(10, *image_shape)
        known_slices = np.arange(1, 11) * 10
        # known_slices = [10, 20, 30, 40, 50, 60, 70, 80, 90, 100]
        gene_data = GeneDataset(
            known_gene, known_slices, "coronal", volume_shape=(110, *image_shape)
        )
        assert gene_data.get_closest_known(10) == 10
        assert gene_data.get_closest_known(110) == 100
        assert gene_data.get_closest_known(78) == 80
        assert gene_data.get_closest_known(55) == 50

    @pytest.mark.parametrize("rgb", [True, False])
    def test_get_surroundings(self, rgb):
        """Test get surrounding gene slices."""
        if rgb:
            image_shape = (20, 30, 3)
        else:
            image_shape = (20, 30)
        known_gene = np.random.rand(10, *image_shape)
        known_slices = np.arange(1, 11) * 10
        # known_slices = [10, 20, 30, 40, 50, 60, 70, 80, 90, 100]
        gene_data = GeneDataset(
            known_gene, known_slices, "coronal", volume_shape=(110, *image_shape)
        )
        assert gene_data.get_surrounding_slices(2) == (None, 10)
        assert gene_data.get_surrounding_slices(10) == (10, 20)
        assert gene_data.get_surrounding_slices(22) == (20, 30)
        assert gene_data.get_surrounding_slices(200) == (100, None)



# Copyright 2021, Blue Brain Project, EPFL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for utility functions."""
from __future__ import annotations


def find_closest(x: float, values: list[int]) -> tuple[int, float]:
    """Find closest element to a given element in a list and its position.

    Parameters
    ----------
    x
        Value to compare with values list.
    values
        List of values where to find closest element to x and its position.

    Returns
    -------
    int
        Indice of the closest element to x in the values list.
    float
        Closest value to x in the values list.
    """
    idx = min(range(len(values)), key=lambda idx: abs(values[idx] - x))

    return idx, values[idx]



# SPDX-License-Identifier: MIT
# Source: https://github.com/microsoft/MaskFlownet/tree/5cba12772e2201f0d1c1e27161d224e585334571
import mxnet as mx
from mxnet.gluon import nn
from mxnet import nd
import numpy as np
from mxnet.base import numeric_types
from mxnet import symbol

class Reconstruction2D(nn.HybridBlock):
	def __init__(self, in_channels = 1, block_grad = False, **kwargs):
		super().__init__(**kwargs)
		self.in_channels = in_channels
		self.block_grad = block_grad

	def hybrid_forward(self, F, x, flow):
		if self.block_grad:
			flow = F.BlockGrad(flow)
		grid = F.GridGenerator(data = flow.flip(axis = 1), transform_type = "warp")
		return F.BilinearSampler(x, grid)

class Reconstruction2DSmooth(nn.HybridBlock):
	def __init__(self, in_channels = 1, block_grad = False, **kwargs):
		super().__init__(**kwargs)
		self.in_channels = in_channels
		self.block_grad = block_grad

	def hybrid_forward(self, F, x, flow):
		if self.block_grad:
			flow = F.BlockGrad(flow)
		grid = F.GridGenerator(data = flow.flip(axis = 1), transform_type = "warp").clip(-1, 1)
		return F.BilinearSampler(x, grid)

class DeformableConv2D(nn.HybridBlock):
	""" Deformable Convolution 2D

	Parameters
	----------
	channels : int
		The dimensionality of the output space
		i.e. the number of output channels in the convolution.
	kernel_size : int or tuple/list of n ints
		Specifies the dimensions of the convolution window.
	strides: int or tuple/list of n ints,
		Specifies the strides of the convolution.
	padding : int or tuple/list of n ints,
		If padding is non-zero, then the input is implicitly zero-padded
		on both sides for padding number of points
	dilation: int or tuple/list of n ints,
		Specifies the dilation rate to use for dilated convolution.
	groups : int
		Controls the connections between inputs and outputs.
		At groups=1, all inputs are convolved to all outputs.
		At groups=2, the operation becomes equivalent to having two convolution
		layers side by side, each seeing half the input channels, and producing
		half the output channels, and both subsequently concatenated.
	layout : str,
		Dimension ordering of data and weight. Can be 'NCW', 'NWC', 'NCHW',
		'NHWC', 'NCDHW', 'NDHWC', etc. 'N', 'C', 'H', 'W', 'D' stands for
		batch, channel, height, width and depth dimensions respectively.
		Convolution is performed over 'D', 'H', and 'W' dimensions.
	in_channels : int, default 0
		The number of input channels to this layer. If not specified,
		initialization will be deferred to the first time `forward` is called
		and `in_channels` will be inferred from the shape of input data.
	activation : str
		Activation function to use. See :func:`~mxnet.ndarray.Activation`.
		If you don't specify anything, no activation is applied
		(ie. "linear" activation: `a(x) = x`).
	use_bias: bool
		Whether the layer uses a bias vector.
	weight_initializer : str or `Initializer`
		Initializer for the `weight` weights matrix.
	bias_initializer: str or `Initializer`
		Initializer for the bias vector.
	"""
	def __init__(self, channels, kernel_size, strides=1, padding=0, dilation=1,
				 groups=1, layout='NCHW', num_deformable_group=1, in_channels=0, activation=None, use_bias=True,
				 weight_initializer=None, bias_initializer='zeros', 
				 prefix=None, params=None):
		super().__init__(prefix=prefix, params=params)
		with self.name_scope():
			self._channels = channels
			self._in_channels = in_channels
			if isinstance(kernel_size, numeric_types):
				kernel_size = (kernel_size,)*2
			if isinstance(strides, numeric_types):
				strides = (strides,)*len(kernel_size)
			if isinstance(padding, numeric_types):
				padding = (padding,)*len(kernel_size)
			if isinstance(dilation, numeric_types):
				dilation = (dilation,)*len(kernel_size)
			self._kwargs = {
				'kernel': kernel_size, 'stride': strides, 'dilate': dilation,
				'pad': padding, 'num_filter': channels, 'num_group': groups,
				'no_bias': not use_bias, 'layout': layout, 
				'num_deformable_group' : num_deformable_group}

			wshapes = [
				(),
				(channels, in_channels) + kernel_size,
				(channels,) 
			]
			self.weight = self.params.get('weight', shape=wshapes[1],
										  init=weight_initializer,
										  allow_deferred_init=True)
			if use_bias:
				self.bias = self.params.get('bias', shape=wshapes[2],
											init=bias_initializer,
											allow_deferred_init=True)
			else:
				self.bias = None

			if activation is not None:
				self.act = nn.Activation(activation, prefix=activation+'_')
			else:
				self.act = None

	def hybrid_forward(self, F, x, offset, weight, bias=None):
		if bias is None:
			act = F.contrib.DeformableConvolution(x, offset, weight, name='fwd', **self._kwargs)
		else:
			act = F.contrib.DeformableConvolution(x, offset, weight, bias, name='fwd', **self._kwargs)
		if self.act is not None:
			act = self.act(act)
		return act

	def _alias(self):
		return 'deformable_conv'

	def __repr__(self):
		s = '{name}({mapping}, kernel_size={kernel}, stride={stride}'
		len_kernel_size = len(self._kwargs['kernel'])
		if self._kwargs['pad'] != (0,) * len_kernel_size:
			s += ', padding={pad}'
		if self._kwargs['dilate'] != (1,) * len_kernel_size:
			s += ', dilation={dilate}'
		if self._kwargs['num_group'] != 1:
			s += ', groups={num_group}'
		if self.bias is None:
			s += ', bias=False'
		s += ')'
		shape = self.weight.shape
		return s.format(name=self.__class__.__name__,
						mapping='{0} -> {1}'.format(shape[1] if shape[1] else None, shape[0]),
						**self._kwargs)



# Copyright 2021, Blue Brain Project, EPFL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Volume interpolation based on pairwise interpolation between slices."""
from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from math import ceil, log2

import numpy as np
import torch
from torchvision.transforms import ToTensor

from atlinter.data import GeneDataset
from atlinter.utils import find_closest


class PairInterpolationModel(ABC):
    """Base class for pair-interpolation models.

    Subclasses of this class implement an interpolation between two given
    images `img1` and `img2` to produce and intermediate image `img_mid`.

    This class and its subclasses are used by the PairInterpolate class,
    which applies a given interpolation model to concrete data.
    """

    def before_interpolation(self, img1, img2):
        """Run initialization and pre-processing steps before interpolation.

        Typical applications of this method are padding and cropping of
        input images to fit the model requirements, as well as initialisation
        of any internal state, should one be necessary.

        Parameters
        ----------
        img1 : np.ndarray
            The left image of shape (width, height)
        img2 : np.ndarray
            The right image of shape (width, height).

        Returns
        -------
        img1 : np.ndarray
            The pre-processed left image.
        img2 : np.ndarray
            The pre-processed right image.
        """
        return img1, img2

    @abstractmethod
    def interpolate(self, img1, img2):
        """Interpolate two images.

        In the typical setting the input images are going to be of the format
        as returned by the `before_interpolation`.

        Parameters
        ----------
        img1 : np.ndarray
            The left image.
        img2 : np.ndarray
            The right image.

        Returns
        -------
        img_mid : np.ndarray
            The interpolated image.
        """

    def after_interpolation(self, interpolated_images):
        """Run any post-processing after all interpolation is done.

        Typical applications are padding and cropping of the image stack,
        as well as any clean-up of the model state.

        Parameters
        ----------
        interpolated_images : np.ndarray
            The stacked interpolated images. The array will include the input
            images as the first and the last items respectively and will
            therefore have the shape (n_interpolated + 2, height, width)

        Returns
        -------
        np.ndarray
            The post-processed interpolated images.
        """
        return interpolated_images


class LinearPairInterpolationModel(PairInterpolationModel):
    """Linear pairwise interpolation.

    This is the simplest possible interpolation model where the middle
    image is the average of the left and right images.
    """

    def interpolate(self, img1, img2):
        """Interpolate two images using linear interpolation.

        Parameters
        ----------
        img1 : np.ndarray
            The left image.
        img2 : np.ndarray
            The right image.

        Returns
        -------
        img_mid : np.ndarray
            The interpolated image.
        """
        img_mid = np.mean([img1, img2], axis=0)

        return img_mid


class RIFEPairInterpolationModel(PairInterpolationModel):
    """Pairwise image interpolation using the RIFE model.

    The typical use is
    >>> from atlinter.vendor.rife.RIFE_HD import Model as RifeModel
    >>> from atlinter.vendor.rife.RIFE_HD import device as rife_device
    >>> rife_model = RifeModel()
    >>> rife_model.load_model("/path/to/train_log", -1)
    >>> rife_model.eval()
    >>> rife_interpolation_model = RIFEPairInterpolationModel(rife_model, rife_device)

    Parameters
    ----------
    rife_model : atlinter.vendor.rife.RIFE_HD.Model
        The RIFE model instance.
    rife_device : from atlinter.vendor.rife.RIFE_HD.device
        The RIFE device.
    """

    def __init__(self, rife_model, rife_device):
        # The behaviour of torch.nn.functional.interpolate has slightly changed,
        # which leads to this warning. It doesn't seem to have an impact on the
        # results, but if the authors of RIFE decide to update their code base
        # by either specifying the `recompute_scale_factor` parameter or by
        # some other means, then this warning filter should be removed.
        # TODO: check the RIFE code for updates and remove the filter if necessary.
        warnings.filterwarnings(
            "ignore",
            "The default behavior for interpolate/upsample with float scale_factor",
            UserWarning,
        )
        self.rife_model = rife_model
        self.rife_device = rife_device
        self.shape = (0, 0)

    def before_interpolation(self, img1, img2):
        """Pad input images to a multiple of 32 pixels.

        Parameters
        ----------
        img1 : np.ndarray
            The left image of shape.
        img2 : np.ndarray
            The right image of shape.

        Returns
        -------
        img1 : np.ndarray
            The padded left image.
        img2 : np.ndarray
            The padded right image.
        """
        image_shape = img1.shape
        if len(image_shape) == 3 and image_shape[-1] == 3:
            rgb = True
            image_shape = image_shape[:-1]
        else:
            rgb = False

        self.shape = np.array(image_shape)
        pad_x, pad_y = ((self.shape - 1) // 32 + 1) * 32 - self.shape

        if rgb:
            img1 = np.pad(img1, ((0, pad_x), (0, pad_y), (0, 0)))
            img2 = np.pad(img2, ((0, pad_x), (0, pad_y), (0, 0)))
        else:
            img1 = np.pad(img1, ((0, pad_x), (0, pad_y)))
            img2 = np.pad(img2, ((0, pad_x), (0, pad_y)))

        return img1, img2

    def interpolate(self, img1, img2):
        """Interpolate two images using RIFE.

        Note: img1 and img2 needs to have the same shape.
        If img1, img2 are grayscale, the dimension should be (height, width).
        If img1, img2 are RGB image, the dimension should be (height, width, 3).

        Parameters
        ----------
        img1 : np.ndarray
            The left image.
        img2 : np.ndarray
            The right image.

        Returns
        -------
        img_mid : np.ndarray
            The interpolated image.
        """
        # Add batch and RGB dimensions (if not already), set device
        if len(img1.shape) == 2:
            rgb = False
            img1 = (
                torch.tensor(img1, dtype=torch.float32)
                .repeat((1, 3, 1, 1))
                .to(self.rife_device)
            )
            img2 = (
                torch.tensor(img2, dtype=torch.float32)
                .repeat((1, 3, 1, 1))
                .to(self.rife_device)
            )
        else:
            rgb = True
            img1 = np.transpose(img1, (2, 0, 1))[np.newaxis]
            img2 = np.transpose(img2, (2, 0, 1))[np.newaxis]
            img1 = torch.tensor(img1, dtype=torch.float32).to(self.rife_device)
            img2 = torch.tensor(img2, dtype=torch.float32).to(self.rife_device)

        # The actual interpolation
        img_mid = self.rife_model.inference(img1, img2).detach().cpu().numpy()

        img_mid = img_mid.squeeze()
        if rgb:
            # Put the RGB channel at the end
            img_mid = np.transpose(img_mid, (1, 2, 0))
        else:
            # Average out the RGB dimension
            img_mid = img_mid.mean(axis=0)

        return img_mid

    def after_interpolation(self, interpolated_images):
        """Undo the padding added in `before_interpolation`.

        Parameters
        ----------
        interpolated_images : np.ndarray
            The stacked interpolated images.
            If input images are grayscale,
            the dimension should be (n_img, height, width) or (height, width).
            If input images are RGB image,
            the dimension should be (n_img, height, width, 3) or (height, width, 3).

        Returns
        -------
        np.ndarray
            The stacked interpolated images with padding removed.
        """
        # No n_img dimension: (height, width) or (height, width, 3)
        if len(interpolated_images.shape) == 2 or (
            len(interpolated_images.shape) == 3 and interpolated_images.shape[-1] == 3
        ):
            return interpolated_images[: self.shape[0], : self.shape[1]]
        # n_img dimension: (n_img, height, width) or (n_img, height, width, 3)
        else:
            return interpolated_images[:, : self.shape[0], : self.shape[1]]


class CAINPairInterpolationModel(PairInterpolationModel):
    """Pairwise image interpolation using the CAIN model.

    The typical use is
    >>> from atlinter.vendor.cain.cain import CAIN
    >>> device = "cuda" if torch.cuda.is_available else "cpu"
    >>> cain_model = torch.nn.DataParallel(CAIN()).to(device)
    >>> cain_checkpoint = torch.load("pretrained_cain.pth", map_location=device)
    >>> cain_model.load_state_dict(cain_checkpoint["state_dict"])
    >>> cain_interpolation_model = CAINPairInterpolationModel(cain_model)

    Parameters
    ----------
    cain_model : atlinter.vendor.cain.cain.CAIN or torch.nn.DataParallel
        The CAIN model instance.
    """

    def __init__(self, cain_model):
        self.cain_model = cain_model
        self.to_tensor = ToTensor()

    def interpolate(self, img1, img2):
        """Interpolate two images using CAIN.

        Note: img1 and img2 needs to have the same shape.
        If img1, img2 are grayscale, the dimension should be (height, width).
        If img1, img2 are RGB image, the dimension should be (height, width, 3).

        Parameters
        ----------
        img1 : np.ndarray
            The left image.
        img2 : np.ndarray
            The right image.

        Returns
        -------
        img_mid : np.ndarray
            The interpolated image.
        """
        # Add batch and RGB dimensions
        if len(img1.shape) == 2:
            rgb = False
            img1 = self.to_tensor(img1).repeat((1, 3, 1, 1))
            img2 = self.to_tensor(img2).repeat((1, 3, 1, 1))
        else:
            rgb = True
            img1 = self.to_tensor(np.transpose(img1, (2, 0, 1)))[None]
            img2 = self.to_tensor(np.transpose(img2, (2, 0, 1)))[None]

        # The actual interpolation
        img_mid, _ = self.cain_model(img1, img2)
        img_mid = img_mid.detach().cpu().numpy()

        img_mid = img_mid.squeeze()
        if rgb:
            # Put the RGB channel at the end
            img_mid = np.transpose(img_mid, (1, 2, 0))
        else:
            # Average out the RGB dimension
            img_mid = img_mid.mean(axis=0)

        return img_mid


class AntsPairInterpolationModel(PairInterpolationModel):
    """Pairwise image interpolation using AntsPy registration.

    Typical use is
    >>> from atlannot.ants import register, transform
    >>> ants_interpolation_model = AntsPairInterpolationModel(register, transform)

    Parameters
    ----------
    register_fn : atlannot.ants.register
        The AntsPy registration function
    transform_fn : atlannot.ants.transform
        The AntsPy transformation function
    """

    def __init__(self, register_fn, transform_fn):
        self.register_fn = register_fn
        self.transform_fn = transform_fn

    def interpolate(self, img1, img2):
        """Interpolate two images using AntsPy registration.

        Parameters
        ----------
        img1 : np.ndarray
            The left image.
        img2 : np.ndarray
            The right image.

        Returns
        -------
        img_mid : np.ndarray
            The interpolated image.
        """
        # Ensure the correct d-type
        img1 = img1.astype(np.float32)
        img2 = img2.astype(np.float32)

        # The actual interpolation
        nii_data = self.register_fn(fixed=img2, moving=img1)
        img_mid = self.transform_fn(img1, nii_data / 2)

        return img_mid


class PairInterpolate:
    """Runner for pairwise interpolation using different models.

    Parameters
    ----------
    n_repeat : int (optional)
        The number of times the interpolation should be iterated. For each
        iteration an interpolated image is inserted between each pair of
        images from the previous iteration. Therefore n_{i+1} = n_i + (n_i + 1).
        For example, for n_repeat=3 the progression of the number of images
        will be the following: input = 0 -> 1 -> 3 -> 7
    """

    def __init__(self, n_repeat=1):
        self.n_repeat = n_repeat

    def repeat(self, n_repeat):
        """Set the number of interpolation iterations.

        Parameters
        ----------
        n_repeat : int
            The new number of interpolation iterations. See `__init__` for more
            details.
        """
        self.n_repeat = n_repeat

        return self

    def __call__(self, img1, img2, model: PairInterpolationModel):
        """Run the interpolation with the given interpolation model.

        Parameters
        ----------
        img1 : np.ndarray
            The left input image.
        img2 : np.ndarray
            The right input image.
        model : PairInterpolationModel
            The interpolation model.

        Returns
        -------
        interpolated_images : np.ndarray
            A stack of interpolation images. The input images are not included
            in this stack.
        """
        img1, img2 = model.before_interpolation(img1, img2)
        interpolated_images = self._repeated_interpolation(
            img1, img2, model, self.n_repeat
        )
        interpolated_images = np.stack(interpolated_images)
        interpolated_images = model.after_interpolation(interpolated_images)

        return interpolated_images

    def _repeated_interpolation(self, img1, img2, model, n_repeat):
        # End of recursion
        if n_repeat <= 0:
            return []

        # Recursion step
        img_mid = model.interpolate(img1, img2)
        left_images = self._repeated_interpolation(img1, img_mid, model, n_repeat - 1)
        right_images = self._repeated_interpolation(img_mid, img2, model, n_repeat - 1)

        return [*left_images, img_mid, *right_images]


class GeneInterpolate:
    """Interpolation of a gene dataset.

    Parameters
    ----------
    gene_data : GeneData
        Gene Dataset to interpolate. It contains a `volume` of reference shape
        with all known places located at the right place and a `metadata` dictionary
        containing information about the axis of the dataset and the section numbers.

    model : PairInterpolationModel
        Pair-interpolation model.
    """

    def __init__(
        self,
        gene_data: GeneDataset,
        model: PairInterpolationModel,
    ):
        self.gene_data = gene_data
        self.model = model

        self.axis = self.gene_data.axis
        self.gene_volume = self.gene_data.volume.copy()
        # If sagittal axis, put the sagittal dimension first
        if self.axis == "sagittal":
            self.gene_volume = np.moveaxis(self.gene_volume, 2, 0)

    def get_interpolation(
        self, left: int, right: int
    ) -> tuple[np.ndarray | None, np.ndarray | None]:
        """Compute the interpolation for a pair of images.

        Parameters
        ----------
        left
            Slice number of the left image to consider.
        right
            Slice number of the right image to consider.

        Returns
        -------
        interpolated_images : np.array or None
            Interpolated image for the given pair of images.
            Array of shape (N, dim1, dim2, 3) with N the number of
            interpolated images.
        predicted_section_numbers : np.array or None
            Slice value of the predicted images.
            Array of shape (N, 1) with N the number of interpolated images.
        """
        diff = right - left
        if diff == 0:
            return None, None

        n_repeat = self.get_n_repeat(diff)

        pair_interpolate = PairInterpolate(n_repeat=n_repeat)
        interpolated_images = pair_interpolate(
            self.gene_volume[left], self.gene_volume[right], self.model
        )
        predicted_section_numbers = self.get_predicted_section_numbers(
            left, right, n_repeat
        )
        return interpolated_images, predicted_section_numbers

    def get_all_interpolation(self) -> tuple[np.ndarray, np.ndarray]:
        """Compute pair interpolation for the entire volume.

        Returns
        -------
        all_interpolated_images : np.array
            Interpolated image for the entire volume.
            Array of shape (N, dim1, dim2, 3) with N the number of
            interpolated images.
        all_predicted_section_numbers : np.array
            Slice value of the predicted images.
            Array of shape (N, 1) with N the number of interpolated images.
        """
        # TODO: Try to change the implementation of the prediction so that
        # we do not predict slices that are not needed.
        known_slices = sorted(self.gene_data.known_slices)

        all_interpolated_images = []
        all_predicted_section_numbers = []
        for i in range(len(known_slices) - 1):
            left, right = known_slices[i], known_slices[i + 1]
            (
                interpolated_images,
                predicted_section_numbers,
            ) = self.get_interpolation(left, right)
            if interpolated_images is None:
                continue
            all_interpolated_images.append(interpolated_images)
            all_predicted_section_numbers.append(predicted_section_numbers)

        all_interpolated_images = np.concatenate(all_interpolated_images)
        all_predicted_section_numbers = np.concatenate(all_predicted_section_numbers)
        return all_interpolated_images, all_predicted_section_numbers

    def predict_slice(self, slice_number: int) -> np.ndarray:
        """Predict one gene slice.

        Parameters
        ----------
        slice_number
            Slice section to predict.

        Returns
        -------
        np.ndarray
            Predicted gene slice. Array of shape (dim1, dim2, 3)
            being (528, 320) for sagittal dataset and
            (320, 456) for coronal dataset.
        """
        left, right = self.gene_data.get_surrounding_slices(slice_number)

        if left is None:
            return self.gene_volume[right]
        elif right is None:
            return self.gene_volume[left]
        else:
            interpolated_images, predicted_section_numbers = self.get_interpolation(
                left, right
            )
            index = find_closest(slice_number, predicted_section_numbers)[0]
            return interpolated_images[index]

    def predict_volume(self) -> np.ndarray:
        """Predict entire volume with known gene slices.

        This function might be slow.
        """
        volume_shape = self.gene_data.volume_shape
        volume = np.zeros(volume_shape)

        if self.gene_data.axis == "sagittal":
            volume = np.moveaxis(volume, 2, 0)
        # Get all the predictions
        (
            all_interpolated_images,
            all_predicted_section_numbers,
        ) = self.get_all_interpolation()

        min_slice_number = min(self.gene_data.known_slices)
        max_slice_number = max(self.gene_data.known_slices)
        end = volume_shape[0] if self.gene_data.axis == "coronal" else volume_shape[2]

        # Populate the volume
        for slice_number in range(end):
            # If the slice is known, just copy the gene.
            if slice_number in self.gene_data.known_slices:
                volume[slice_number] = self.gene_volume[slice_number]
            # If the slice section is smaller than all known slice
            # We copy-paste the smallest known slice.
            elif slice_number < min_slice_number:
                volume[slice_number] = self.gene_volume[min_slice_number]
            # If the slice section is bigger than all known slice
            # We copy-paste the biggest known slice.
            elif slice_number > max_slice_number:
                volume[slice_number] = self.gene_volume[max_slice_number]
            # If the slice is surrounded by two known slice.
            # Determine the prediction closest to the slice section.
            else:
                index = find_closest(slice_number, all_predicted_section_numbers)[0]
                volume[slice_number] = all_interpolated_images[index]

        if self.gene_data.axis == "sagittal":
            volume = np.moveaxis(volume, 0, 2)

        return volume

    @staticmethod
    def get_n_repeat(diff: int) -> int:
        """Determine the number of repetitions to compute."""
        if diff <= 0:
            return 0
        n_repeat = ceil(log2(diff))
        return n_repeat

    @staticmethod
    def get_predicted_section_numbers(
        left: int, right: int, n_repeat: int
    ) -> np.ndarray:
        """Get slice values of predicted images."""
        n_steps = 2 ** n_repeat + 1
        predicted_section_numbers = np.linspace(left, right, n_steps)
        return predicted_section_numbers[1:-1]



# SPDX-License-Identifier: MIT
# Source: https://github.com/hzwer/arXiv2020-RIFE/tree/6ff174584737a9aa27cd9654443a8a79c76799c9
import torch
import torch.nn as nn
import torch.nn.functional as F
from .warplayer import warp

def deconv(in_planes, out_planes, kernel_size=4, stride=2, padding=1):
    return nn.Sequential(
        torch.nn.ConvTranspose2d(in_channels=in_planes, out_channels=out_planes, kernel_size=4, stride=2, padding=1),
        nn.PReLU(out_planes)
    )

def conv_wo_act(in_planes, out_planes, kernel_size=3, stride=1, padding=1, dilation=1):
    return nn.Sequential(
        nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride,
                  padding=padding, dilation=dilation, bias=True),
        )

def conv(in_planes, out_planes, kernel_size=3, stride=1, padding=1, dilation=1):
    return nn.Sequential(
        nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride,
                  padding=padding, dilation=dilation, bias=True),
        nn.PReLU(out_planes)
    )

class IFBlock(nn.Module):
    def __init__(self, in_planes, scale=1, c=64):
        super(IFBlock, self).__init__()
        self.scale = scale
        self.conv0 = nn.Sequential(
            conv(in_planes, c//2, 3, 2, 1),
            conv(c//2, c, 3, 2, 1),
            )
        self.convblock = nn.Sequential(
            conv(c, c),
            conv(c, c),
            conv(c, c),
            conv(c, c),
            conv(c, c),
            conv(c, c),
            conv(c, c),
            conv(c, c),
        )
        self.conv1 = nn.ConvTranspose2d(c, 4, 4, 2, 1)

    def forward(self, x):
        if self.scale != 1:
            x = F.interpolate(x, scale_factor= 1. / self.scale, mode="bilinear", align_corners=False)
        x = self.conv0(x)
        x = self.convblock(x) + x
        x = self.conv1(x)
        flow = x
        if self.scale != 1:
            flow = F.interpolate(flow, scale_factor= self.scale, mode="bilinear", align_corners=False)
        return flow
    
class IFNet(nn.Module):
    def __init__(self):
        super(IFNet, self).__init__()
        self.block0 = IFBlock(6, scale=4, c=320)
        self.block1 = IFBlock(10, scale=2, c=225)
        self.block2 = IFBlock(10, scale=1, c=135)

    def forward(self, x):
        flow0 = self.block0(x)
        F1 = flow0
        F1_large = F.interpolate(F1, scale_factor=2.0, mode="bilinear", align_corners=False) * 2.0
        warped_img0 = warp(x[:, :3], F1_large[:, :2])
        warped_img1 = warp(x[:, 3:], F1_large[:, 2:4])
        flow1 = self.block1(torch.cat((warped_img0, warped_img1, F1_large), 1))
        F2 = (flow0 + flow1)
        F2_large = F.interpolate(F2, scale_factor=2.0, mode="bilinear", align_corners=False) * 2.0
        warped_img0 = warp(x[:, :3], F2_large[:, :2])
        warped_img1 = warp(x[:, 3:], F2_large[:, 2:4])
        flow2 = self.block2(torch.cat((warped_img0, warped_img1, F2_large), 1))
        F3 = (flow0 + flow1 + flow2)
        return F3, [F1, F2, F3]



# Copyright 2021, Blue Brain Project, EPFL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Gene dataset module."""
from __future__ import annotations

import numpy as np

from atlinter.utils import find_closest


class GeneDataset:
    """Class representing gene dataset.

    Parameters
    ----------
    gene
        Array containing all gene slices known from a given dataset.
        Dimensions should correspond to (n_genes, dim1, dim2)
        for grayscale images and (n_genes, dim1, dim2, 3) for RGB.
    section_numbers
        List of the section numbers of the gene slices.
    axis
        Axis of the gene dataset.
    volume_shape
        Shape of the final volume.
        Dimensions should correspond to (coronal, transverse, sagittal)
        for grayscale images and (coronal, transverse, sagittal, 3) for RGB.
    """

    def __init__(
        self,
        gene: np.ndarray,
        section_numbers: list[int],
        axis: str,
        volume_shape: tuple[int, int, int]
        | tuple[int, int, int, int] = (528, 320, 456, 3),
    ):
        """Instantiate gene dataset."""
        if not isinstance(gene, np.ndarray) or gene.ndim not in {3, 4}:
            raise ValueError(
                "The gene has to be an array of \n"
                "* 3 dimensions (n_genes, dim1, dim2) for grayscale images. \n"
                "* 4 dimensions (n_genes, dim1, dim2, 3) for RBG images."
            )
        self.gene = gene
        self.known_slices = section_numbers
        self.axis = axis
        self.volume_shape = volume_shape

        # Create volume
        self.volume = np.zeros(volume_shape)
        end = volume_shape[0] if self.axis == "coronal" else volume_shape[2]
        slices = [s for s in self.known_slices if s < end]

        # Populate volume with known gene slices
        if self.axis == "coronal":
            self.volume[slices] = self.gene
        elif self.axis == "sagittal":
            self.volume[:, :, slices] = np.moveaxis(self.gene, 0, 2)

        # Sorting known slices
        self.known_slices = sorted(self.known_slices)
        # Compute the unknown slices
        all_slices = np.arange(end)
        self.unknown_slices = [int(s) for s in all_slices if s not in self.known_slices]

    def get_closest_known(self, slice_number: int) -> int:
        """Obtain closest gene slice known for a given slice number."""
        return find_closest(slice_number, self.known_slices)[1]

    def get_surrounding_slices(
        self, slice_number: int
    ) -> tuple[int | None, int | None]:
        """Get the two known surrounding gene slices for a given slice number.

        If index is smaller than all indices, the result will be (None, smallest indice)
        If index is bigger than all indices, the result will be (biggest indice, None)
        If index is equal to one of the indices, the result will be (index, next indice)

        Parameters
        ----------
        slice_number : int
            Index to look for surrounding gene slices

        Returns
        -------
        left : int
            Left value to the index among indices.
        right : int
            Right value to the index among indices.
        """
        # Special cases
        if slice_number < self.known_slices[0]:
            return None, self.known_slices[0]
        if slice_number >= self.known_slices[-1]:
            return self.known_slices[-1], None

        # Find the first index that is strictly bigger than the given one.
        # After the special cases the indices list is at least of length 2.
        right_pos = 0
        while slice_number >= self.known_slices[right_pos]:
            right_pos += 1

        return self.known_slices[right_pos - 1], self.known_slices[right_pos]



# Copyright 2021, Blue Brain Project, EPFL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Align ISH gene to Nissl slices."""
import argparse
import json
import pathlib
import sys

import numpy as np
from atlannot.ants import register, transform
from skimage.color import rgb2gray


def parse_args():
    """Parse command line arguments.

    Returns
    -------
    args : argparse.Namespace
        The parsed command line arguments.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("gene_name")
    args = parser.parse_args()

    return args


def dataset_registration(nissl, gene, section_numbers, axis):
    """Align ISH data to a given nissl slice thanks to ANTsPY.

    Parameters
    ----------
    nissl : np.ndarray
        Corresponding nissl section for the given gene.
    gene : np.ndarray
        Dataset of gene expression images to align to Nissl.
    section_numbers : list
        List of section numbers of the corresponding genes
    axis : {'coronal', 'sagittal'}
        Axis of the gene expressions

    Returns
    -------
    aligned_gene : np.ndarray
        Aligned gene expressions.
    """
    aligned_gene = []
    for single_gene, section_number in zip(gene, section_numbers):
        if axis == "coronal":
            try:
                single_nissl = nissl[int(section_number), :, :]
            except IndexError:
                aligned_gene.append(single_gene)
                continue
        elif axis == "sagittal":
            single_nissl = nissl[:, :, int(section_number)]
        else:
            raise ValueError(
                f"Axis {axis} is not supported"
                "Only coronal and sagittal are supported!"
            )

        aligned_gene += [
            single_registration(single_nissl, single_gene),
        ]

    aligned_gene = np.array(aligned_gene)
    return aligned_gene


def single_registration(single_nissl, single_gene):
    """Align ISH data to a given nissl slice thanks to ANTsPY.

    Parameters
    ----------
    single_nissl : np.ndarray
        Corresponding nissl section for the given gene.
    single_gene : np.ndarray
        Gene expression to align to Nissl.

    Returns
    -------
    aligned_single_gene : np.ndarray
        Aligned gene expression.
    """
    rgb = False
    tmp_gene = single_gene.copy()

    if len(single_gene.shape) == 3:
        rgb = True
        tmp_gene = 1 - rgb2gray(tmp_gene).astype(np.float32)

    if single_nissl.shape != tmp_gene.shape:
        raise ValueError(
            f"Nissl ({single_nissl.shape}) and gene "
            f"({tmp_gene.shape}) should have the same shape"
        )

    nii_data = register(single_nissl, tmp_gene)

    if rgb:
        aligned_single_gene = np.zeros_like(single_gene)
        aligned_single_gene[:, :, 0] = transform(
            single_gene[:, :, 0].astype(np.float32),
            nii_data,
            defaultvalue=single_gene[0, 0, 0],
        )
        aligned_single_gene[:, :, 1] = transform(
            single_gene[:, :, 1].astype(np.float32),
            nii_data,
            defaultvalue=single_gene[0, 0, 1],
        )
        aligned_single_gene[:, :, 2] = transform(
            single_gene[:, :, 2].astype(np.float32),
            nii_data,
            defaultvalue=single_gene[0, 0, 2],
        )
    else:
        aligned_single_gene = transform(single_gene, nii_data)

    return aligned_single_gene


def main():
    """Align gene expression to Nissl."""
    # Imports
    import nrrd

    args = parse_args()
    gene_name = args.gene_name

    cwd = pathlib.Path(__file__).parent.resolve()
    nissl_path = cwd / "ara_nissl_25.nrrd"
    if not nissl_path.exists():
        raise ValueError("Nissl volume is missing! Please dvc pull download-nissl")

    nissl, _ = nrrd.read(nissl_path)
    nissl = nissl / nissl.max()

    for axis in ["sagittal", "coronal"]:

        file_dir = cwd / axis / gene_name
        if not file_dir.exists():
            continue
        experiment_ids = [
            path.stem for path in file_dir.iterdir() if path.suffix == ".npy"
        ]

        saving_dir = cwd / "aligned" / axis / gene_name
        if not saving_dir.exists():
            saving_dir.mkdir(parents=True)

        for experiment in experiment_ids:
            # Loading
            gene = np.load(file_dir / f"{experiment}.npy")
            with open(file_dir / f"{experiment}.json") as f:
                metadata = json.load(f)
            section_numbers = metadata["section_numbers"]

            # Registration
            aligned_gene = dataset_registration(nissl, gene, section_numbers, axis)

            # Saving
            np.save(saving_dir / f"{experiment}.npy", aligned_gene)
            with open(saving_dir / f"{experiment}.json", "w") as f:
                json.dump(metadata, f)


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2021, Blue Brain Project, EPFL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Optical flow computation based on pair of images."""
import argparse
import logging
from abc import ABC, abstractmethod

import mxnet as mx
import numpy as np
import torch
from scipy import ndimage

from atlinter.data import GeneDataset

logger = logging.getLogger(__name__)


class OpticalFlow(ABC):
    """Class representing optical flow model."""

    def preprocess_images(self, img1, img2):
        """Preprocess image.

        Parameters
        ----------
        img1 : np.ndarray
            The left image.
        img2 : np.ndarray
            The right image.

        Returns
        -------
        img1 : np.ndarray
            The pre-processed left image.
        img2 : np.ndarray
            The pre-processed right image.
        """
        return img1, img2

    @abstractmethod
    def predict_flow(self, img1, img2):
        """Compute optical flow between two images.

        Parameters
        ----------
        img1 : np.ndarray
            The left image.
        img2 : np.ndarray
            The right image.

        Returns
        -------
        flow : np.ndarray
            The optical flow of shape ``(*image.shape, 2)``.
        """

    @classmethod
    def warp_image(cls, flow, img2, order=1):
        """Warp image with the predicted flow.

        Parameters
        ----------
        flow : np.ndarray
            The predicted optical flow of shape ``(*image.shape, 2)``.
        img2 : np.ndarray
            The right image.
        order : int
            The interpolation order. 0 = nearest neighbour, 1 = linear,
            2 = cubic, etc.

        Returns
        -------
        warped : np.ndarray
            The warped image.
        """
        if flow.shape[1:3] != img2.shape[0:2]:
            raise ValueError("The flow shape and the image shape should be consistent.")

        if img2.ndim == 2:
            # greyscale
            ni, nj = img2.shape
            ij = np.mgrid[:ni, :nj]
            warped = ndimage.map_coordinates(img2, ij - flow, order=order)
        elif img2.ndim == 3 and img2.shape[-1] == 3:
            # RGB, warp each channel separately
            warped = np.stack(
                [
                    cls.warp_image(flow, img2[..., 0], order=order),
                    cls.warp_image(flow, img2[..., 1], order=order),
                    cls.warp_image(flow, img2[..., 2], order=order),
                ],
                axis=-1,
            )
        else:
            raise ValueError(f"Invalid image shape: {img2.shape}")

        return warped


class MaskFlowNet(OpticalFlow):
    """MaskFlowNet model for optical flow computation.

    The typical use is

    >>> from atlinter.optical_flow import MaskFlowNet
    >>> checkpoint = "data/checkpoints/maskflownet.params"
    >>> # Make sure the checkpoint exists and uncomment the following line
    >>> # net = MaskFlowNet(checkpoint)

    Parameters
    ----------
    checkpoint_path : str
        Path to checkpoint of the model. See references for possibilities.
    gpu_device : str
        Device to load optical flow model.
    """

    from atlinter.vendor.MaskFlowNet import network, predict_new_data

    def __init__(self, checkpoint_path, gpu_device=""):
        config_file = {
            "network": {"class": "MaskFlownet"},
            "optimizer": {
                "learning_rate": [
                    [300000, 0.0001],
                    [500000, 5e-05],
                    [600000, 2.5e-05],
                    [700000, 1.25e-05],
                    [800000, 6.25e-06],
                ],
                "wd": 0.0004,
            },
        }
        config = self.network.config.Reader(config_file)

        if gpu_device == "":
            ctx = [mx.cpu()]
        else:
            ctx = [mx.gpu(int(gpu_id)) for gpu_id in gpu_device.split(",")]

        pipe = self.network.get_pipeline("MaskFlownet", ctx=ctx, config=config)
        self.pipe = self.predict_new_data.load_checkpoint(pipe, config, checkpoint_path)

    def preprocess_images(self, img1, img2):
        """Preprocess image.

        Parameters
        ----------
        img1 : np.ndarray
            The left image.
        img2 : np.ndarray
            The right image.

        Returns
        -------
        img1 : np.ndarray
            The pre-processed left image.
        img2 : np.ndarray
            The pre-processed right image.
        """
        if img1.shape != img2.shape:
            raise ValueError("The two images have not the same shape!")

        if np.max(img1) <= 1:
            img1 = (img1 * 255).astype(np.uint8)
            img2 = (img2 * 255).astype(np.uint8)

        if len(img1.shape) == 2:
            img1 = np.repeat(img1[..., np.newaxis], 3, axis=2)
            img2 = np.repeat(img2[..., np.newaxis], 3, axis=2)

        return img1, img2

    def predict_flow(self, img1, img2):
        """Compute optical flow between two images.

        Parameters
        ----------
        img1 : np.ndarray
            The left image.
        img2 : np.ndarray
            The right image.

        Returns
        -------
        flow : np.ndarray
            The optical flow.
        """
        flow, _, _ = self.predict_new_data.predict_image_pair_flow(
            img1, img2, self.pipe
        )
        return np.transpose(flow, (2, 0, 1))


class RAFTNet(OpticalFlow):
    """RAFT model for optical flow computation.

    The typical use is

    >>> from atlinter.optical_flow import RAFTNet
    >>> checkpoint = "data/checkpoints/RAFT/models/raft-things.pth"
    >>> # Make sure the checkpoint exists and uncomment the following line
    >>> # net = RAFTNet(checkpoint)

    Parameters
    ----------
    path : str
        Path to the RAFT model.
    device : {'cpu', 'cuda'}
        Device to load optical flow model.
    """

    from atlinter.vendor.RAFT.raft import RAFT

    def __init__(self, path, device="cpu"):
        self.device = device
        args = self.initialize_namespace()
        self.model = torch.nn.DataParallel(self.RAFT(args))
        self.model.load_state_dict(torch.load(path, map_location=torch.device(device)))
        self.model = self.model.module
        if self.device == "cuda":
            self.model.to(self.device)
        self.model.eval()

    @staticmethod
    def initialize_namespace():
        """Initialize namespace needed for RAFT initialization.

        Returns
        -------
        args : argparse.Namespace
            Arguments needed to instantiate RAFT model.
        """
        namespace = argparse.Namespace()
        namespace.small = False
        namespace.mixed_precision = False
        return namespace

    def preprocess_images(self, img1, img2):
        """Preprocess image.

        Parameters
        ----------
        img1 : np.ndarray
            The left image.
        img2 : np.ndarray
            The right image.

        Returns
        -------
        img1 : np.ndarray
            The pre-processed left image.
        img2 : np.ndarray
            The pre-processed right image.
        """
        if img1.shape != img2.shape:
            raise ValueError("The two images have not the same shape!")

        # The image dimensions need to be divisible by 8
        self.shape = np.array(img1.shape[:2])
        pad_x, pad_y = ((self.shape - 1) // 8 + 1) * 8 - self.shape

        if np.max(img1) <= 1:
            img1 = (img1 * 255).astype(np.uint8)
            img2 = (img2 * 255).astype(np.uint8)

        if len(img1.shape) == 2:
            img1 = np.pad(img1, ((0, pad_x), (0, pad_y)))
            img2 = np.pad(img2, ((0, pad_x), (0, pad_y)))

            img1 = np.repeat(img1[np.newaxis], 3, axis=0)
            img2 = np.repeat(img2[np.newaxis], 3, axis=0)
        else:
            img1 = np.pad(img1, ((0, pad_x), (0, pad_y), (0, 0)))
            img2 = np.pad(img2, ((0, pad_x), (0, pad_y), (0, 0)))

            img1 = np.transpose(img1, (2, 0, 1))
            img2 = np.transpose(img2, (2, 0, 1))

        img1 = img1[np.newaxis]
        img2 = img2[np.newaxis]

        return img1, img2

    def predict_flow(self, img1, img2):
        """Compute optical flow between two images.

        Parameters
        ----------
        img1 : np.ndarray
            The left image.
        img2 : np.ndarray
            The right image.

        Returns
        -------
        flow : np.ndarray
            The optical flow.
        """
        img1 = torch.from_numpy(img1).float().to(self.device)
        img2 = torch.from_numpy(img2).float().to(self.device)

        _, flow_up = self.model(img1, img2, iters=20, test_mode=True)
        if self.device == "cuda":
            flow_up = flow_up.cpu()
        flow = flow_up.detach().numpy()[0]
        flow = flow[:, : self.shape[0], : self.shape[1]]
        return flow


class GeneOpticalFlow:
    """Computation of optical flow for gene dataset.

    Parameters
    ----------
    gene_data
        Gene Dataset. It contains a ``volume`` of reference shape
        with all known slices located at the right place and a ``metadata``
        dictionary containing information about the axis of the dataset and the
        section numbers.
    reference_volume
        Reference volume used to compute optical flow. It needs to be of same
        shape as ``gene_data.volume``.
    model
        Model computing flow between two images.
    """

    def __init__(
        self, gene_data: GeneDataset, reference_volume: np.ndarray, model: OpticalFlow
    ):
        self.gene_data = gene_data
        self.reference_volume = reference_volume
        self.model = model

        self.axis = self.gene_data.axis

        if self.axis == "coronal":
            self.gene_volume = self.gene_data.volume.copy()
            self.reference_volume = reference_volume
        elif self.axis == "sagittal":
            self.gene_volume = np.moveaxis(self.gene_data.volume, 2, 0)
            self.reference_volume = np.moveaxis(reference_volume, 2, 0)
        else:
            raise ValueError(f"Unknown axis: {self.axis}")

        # Use to save all flow predictions
        self.pairwise_ref_flows = {}

    def predict_ref_flow(self, idx_from: int, idx_to: int) -> np.ndarray:
        """Compute optical flow between two given slices of the reference volume.

        Parameters
        ----------
        idx_from
            First section to consider
        idx_to
            Second section to consider.

        Returns
        -------
        flow : np.ndarray
            Predicted flow between the two given sections of the reference
            volume.

        Raises
        ------
        ValueError
            If one of the ``idx_from`` and ``idx_to`` is out of the boundaries
            of the reference space.
        """
        n_slices = len(self.gene_volume)
        if not (0 <= idx_from < n_slices and 0 <= idx_to < n_slices):
            raise ValueError(
                f"Slices ({idx_from} and {idx_to})"
                f"have to be between 0 and {n_slices}"
            )

        if (idx_from, idx_to) in self.pairwise_ref_flows:
            return self.pairwise_ref_flows[(idx_from, idx_to)]

        preimg1, preimg2 = self.model.preprocess_images(
            self.reference_volume[idx_from], self.reference_volume[idx_to]
        )
        flow = self.model.predict_flow(preimg1, preimg2)
        self.pairwise_ref_flows[(idx_from, idx_to)] = flow
        return flow

    def predict_slice(self, slice_number: int) -> np.ndarray:
        """Predict one gene slice.

        Parameters
        ----------
        slice_number
            Slice section to predict.

        Returns
        -------
        np.ndarray
            Predicted gene slice. Array of shape ``(dim1, dim2, 3)``
            being ``(528, 320)`` for sagittal dataset and
            ``(320, 456)`` for coronal dataset.
        """
        closest = self.gene_data.get_closest_known(slice_number)
        flow = self.predict_ref_flow(slice_number, closest)

        closest_slice = self.gene_volume[closest]
        predicted_slice = self.model.warp_image(flow, closest_slice)

        return predicted_slice

    def predict_volume(self) -> np.ndarray:
        """Predict entire volume with known gene slices.

        This function might be slow.

        Returns
        -------
        np.ndarray
            Entire gene volume. Array of shape of the volume ``GeneDataset``.
        """
        volume_shape = self.gene_volume.shape
        volume = np.zeros(volume_shape)

        # Populate the volume
        for slice_number in range(volume.shape[0]):
            # If the slice is known, just copy the gene.
            if slice_number in self.gene_data.known_slices:
                volume[slice_number] = self.gene_volume[slice_number]
            # If the slice is unknown, predict it
            else:
                volume[slice_number] = self.predict_slice(slice_number)

        if self.gene_data.axis == "sagittal":
            volume = np.moveaxis(volume, 0, 2)

        return volume



# Copyright 2021, Blue Brain Project, EPFL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unittest import mock

import numpy as np
import torch

from atlinter.metrics import FID
from atlinter.metrics.fid import InceptionV3


@mock.patch("atlinter.metrics.fid.InceptionV3", spec=InceptionV3)
def test_fid(inception_class):
    # Input data
    n_images = 10
    images_1 = np.random.rand(n_images, 5, 5)
    images_2 = np.random.rand(n_images, 5, 5)

    # Mock for the inception class instance and its forward pass
    activations = torch.rand((n_images, 10, 1, 1), dtype=torch.float32)
    inception_instance = inception_class.return_value
    inception_instance.to.return_value = inception_instance
    inception_instance.eval.return_value = inception_instance
    inception_instance.return_value = [activations]

    # Test device is set automatically
    fid = FID()
    assert fid.device is not None
    assert fid.device in {"cuda", "cpu"}

    # Test score computation
    fid = FID(device="cpu")
    score = fid.score(images_1, images_2)
    assert inception_instance.call_count == 2
    assert isinstance(score, float)

    # Test two identical stacks of images have FID=0
    score = fid.score(images_1, images_1)
    assert np.allclose(score, 0)



# Copyright 2021, Blue Brain Project, EPFL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Download special dataset (with 2 genes) from Allen Brain."""
import argparse
import pathlib
import sys

import requests


def parse_args():
    """Parse command line arguments.

    Returns
    -------
    args : argparse.Namespace
        The parsed command line arguments.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("dataset_id")
    args = parser.parse_args()

    return args


def allen_dataset_info(dataset_id):
    """Extract dataset info from Allen API.

    Parameters
    ----------
    dataset_id: int or str
        Dataset ID from Allen Brain API

    Returns
    -------
    dataset_info: dict
        Dictionary from Allen Brain containing
        some metadata for the given dataset.
    """
    url = "http://api.brain-map.org/api/v2/data/query.json?"
    url += f"criteria=model::SectionDataSet,rma::criteria,[id$eq{dataset_id}]"
    r = requests.get(url)
    raw = r.json()
    dataset_info = raw["msg"][0]
    return dataset_info


def main():
    """Download gene expression dataset."""
    # Imports
    import json

    import numpy as np
    import PIL
    from atldld.sync import download_dataset

    args = parse_args()
    # To avoid DecompressionBombError --> highest value from 3 datasets we are using
    PIL.Image.MAX_IMAGE_PIXELS = 192491520

    # Download dataset on allen
    dataset_id = args.dataset_id
    dataset_info = allen_dataset_info(dataset_id)
    dataset_info["dataset_id"] = dataset_id
    dataset = list(download_dataset(dataset_id))
    dataset_info["num_images"] = len(dataset)
    # dataset = sequence of (image_id, p, img, df)

    # Create volume with gene expression
    p_values = [data[1] for data in dataset]
    dataset_info["original_p_values"] = p_values
    p_values = [p // 25 for p in p_values]

    image_ids = [int(data[0]) for data in dataset]
    dataset_info["image_ids"] = image_ids
    # Compute the shape of the output volume based on the shape of
    # the first slice and the highest p-value
    _, _, img_0, df_0 = dataset[0]
    img_0_reg = df_0.warp(img_0)
    shape = (int(max(p_values) + 1), *img_0_reg.shape[:2])

    red_gene = np.zeros(shape)
    green_gene = np.zeros(shape)
    blue_gene = np.zeros(shape)
    all_dfs = []
    for _, p, img, df in dataset:
        section_number = int(p // 25)
        img_reg = df.warp(img)
        red_gene[section_number, :, :] = img_reg[:, :, 0]
        green_gene[section_number, :, :] = img_reg[:, :, 1]
        blue_gene[section_number, :, :] = img_reg[:, :, 2]
        all_dfs.append(df)

    output_dir = pathlib.Path("special_volumes") / f"dataset_{dataset_id}"
    output_dir.mkdir(parents=True)
    np.save(output_dir / dataset_info["red_channel"], red_gene)
    np.save(output_dir / dataset_info["green_channel"], green_gene)
    np.save(output_dir / dataset_info["blue_channel"], blue_gene)
    np.save(output_dir / "dfs.npy", np.stack(all_dfs))
    with open(output_dir / "metadata.json", "w") as f:
        json.dump(dataset_info, f)


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2021, Blue Brain Project, EPFL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The setup script."""
from setuptools import find_packages, setup

with open("README.md", encoding="utf-8") as fh:
    long_description = fh.read()

install_requires = [
    "mxnet",
    "numpy",
    "pytorch-fid",
    "pyyaml",
    "scipy",
    "torch",
    "torchvision",
]
extras_require = {
    "data": [
        "atlannot",
        "atldld==0.2.2",
        "dvc[ssh]",
        "pillow",
        "pynrrd",
        "requests",
        "scikit-image",
    ],
    "dev": [
        "bandit",
        "black",
        "flake8",
        "flake8-bugbear",
        "flake8-comprehensions",
        "flake8-docstrings",
        "isort",
        "py-vendor==0.1.2",
        "pytest",
        "pytest-cov",
        "tox",
    ],
    "docs": [
        "sphinx",
        "sphinx-bluebrain-theme",
    ],
    "optical": [
        "opencv-python",
        "flow_vis",
        "moviepy",
    ],
}

setup(
    name="atlinter",
    author="Blue Brain Project, EPFL",
    license="Apache-2.0",
    use_scm_version={
        "write_to": "src/atlinter/version.py",
        "write_to_template": '"""The package version."""\n__version__ = "{version}"\n',
        "local_scheme": "no-local-version",
    },
    description="Interpolate missing section images in gene expression volumes",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/BlueBrain/atlas-interpolation",
    project_urls={
        "Source": "https://github.com/BlueBrain/atlas-interpolation",
        "Tracker": "https://github.com/BlueBrain/atlas-interpolation/issues",
    },
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: Unix",
        "Operating System :: MacOS",
        "Programming Language :: Python :: 3.7",
    ],
    package_dir={"": "src"},
    packages=find_packages("src"),
    python_requires="~=3.7.0",
    install_requires=install_requires,
    extras_require=extras_require,
)



# Copyright 2021, Blue Brain Project, EPFL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import atlinter


def test_version_exists():
    # Version exists
    assert hasattr(atlinter, "__version__")
    assert isinstance(atlinter.__version__, str)
    parts = atlinter.__version__.split(".")

    # Version has correct format
    # allow for an optional ".devXX" part for local testing
    assert len(parts) in {3, 4}
    assert parts[0].isdecimal()  # major
    assert parts[1].isdecimal()  # minor
    assert parts[2].isdecimal()  # patch
